﻿using System.Text;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.DotProlog.Code;
using OpcLabs.DotProlog.Execution;
using OpcLabs.DotProlog.InputOutput;
using OpcLabs.DotProlog.Terms;
using OpcLabs.DotProlog.Test;

namespace DotPrologTest.Execution
{
    [TestClass]
    // ReSharper disable once InconsistentNaming
    public class Cut_0_Test : ExecutionTest
    {
        #region Static Private Methods

        // ReSharper disable once SuggestBaseTypeForParameter
        static private void Consult([NotNull] IProgram program)
        {
            IDatabase database = program.Database;

            database.Assert(ParseClause("twice(!) :- write('C ')."));
            database.Assert(ParseClause("twice(true) :- write('Moss ')."));

            database.Assert(ParseClause("goal((twice(_), !))."));
            database.Assert(ParseClause("goal(write('Three '))."));
        }

        #endregion



        #region Test Methods
        // ReSharper disable InconsistentNaming

        [TestMethod]
        public void Test_7_8_4_4_Examples_1()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);
                Test(engine, Parse("!. "), Success(), "Expected success.");
            }
        }

        [TestMethod]
        public void Test_7_8_4_4_Examples_2()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);
                Test(engine, Parse("(!, fail; true). "), Failure(), "Expected failure.");
            }
        }

        [TestMethod]
        public void Test_7_8_4_4_Examples_3()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);
                Test(engine, Parse("(call(!), fail; true). "), Success(), "Expected success.");
            }
        }

        [TestMethod]
        public void Test_7_8_4_4_Examples_4()
        {
            var stringBuilder = new StringBuilder();
            using (IEngine engine = new Engine { Environment = { CurrentOutputStream = TextStream.Create(stringBuilder) } })
            {
                Consult(engine.Program);

                Term goal = Parse("twice(_), !, write('Forwards '), fail. ");
                Test(engine, goal, Failure(), "Expected failure.");
                Assert.AreEqual("C Forwards ", stringBuilder.ToString());
            }
        }

        [TestMethod]
        public void Test_7_8_4_4_Examples_5()
        {
            var stringBuilder = new StringBuilder();
            using (IEngine engine = new Engine { Environment = { CurrentOutputStream = TextStream.Create(stringBuilder) } })
            {
                Consult(engine.Program);

                Term goal = Parse("(! ; write('No ')), write('Cut disjunction '), fail. ");
                Test(engine, goal, Failure(), "Expected failure.");
                Assert.AreEqual("Cut disjunction ", stringBuilder.ToString());
            }
        }

        [TestMethod]
        public void Test_7_8_4_4_Examples_6()
        {
            var stringBuilder = new StringBuilder();
            using (IEngine engine = new Engine { Environment = { CurrentOutputStream = TextStream.Create(stringBuilder) } })
            {
                Consult(engine.Program);

                Term goal = Parse("twice(_), (write('No '); !), write('Cut '), fail. ");
                Test(engine, goal, Failure(), "Expected failure.");
                Assert.AreEqual("C No Cut Cut ", stringBuilder.ToString());
            }
        }

        [TestMethod]
        public void Test_7_8_4_4_Examples_7()
        {
            var stringBuilder = new StringBuilder();
            using (IEngine engine = new Engine { Environment = { CurrentOutputStream = TextStream.Create(stringBuilder) } })
            {
                Consult(engine.Program);

                Term goal = Parse("twice(_), (!, fail; write('No ')). ");
                Test(engine, goal, Failure(), "Expected failure.");
                Assert.AreEqual("C ", stringBuilder.ToString());
            }
        }

        [TestMethod]
        public void Test_7_8_4_4_Examples_8()
        {
            var stringBuilder = new StringBuilder();
            using (IEngine engine = new Engine { Environment = { CurrentOutputStream = TextStream.Create(stringBuilder) } })
            {
                Consult(engine.Program);

                Term goal = Parse("twice(X), call(X), write('Forwards '), fail. ");
                Test(engine, goal, Failure(), "Expected failure.");
                Assert.AreEqual("C Forwards Moss Forwards ", stringBuilder.ToString());
            }
        }

        [TestMethod]
        public void Test_7_8_4_4_Examples_9()
        {
            var stringBuilder = new StringBuilder();
            using (IEngine engine = new Engine { Environment = { CurrentOutputStream = TextStream.Create(stringBuilder) } })
            {
                Consult(engine.Program);

                Term goal = Parse("goal(X), call(X), write('Forwards '), fail. ");
                Test(engine, goal, Failure(), "Expected failure.");
                Assert.AreEqual("C Forwards Three Forwards ", stringBuilder.ToString());
            }
        }

        [TestMethod]
        public void Test_7_8_4_4_Examples_10()
        {
            var stringBuilder = new StringBuilder();
            using (IEngine engine = new Engine { Environment = { CurrentOutputStream = TextStream.Create(stringBuilder) } })
            {
                Consult(engine.Program);

                Term goal = Parse(@"twice(_), \+(\+(!)), write('Forwards '), fail. ");
                Test(engine, goal, Failure(), "Expected failure.");
                Assert.AreEqual("C Forwards Moss Forwards ", stringBuilder.ToString());
            }
        }

        [TestMethod]
        public void Test_7_8_4_4_Examples_11()
        {
            var stringBuilder = new StringBuilder();
            using (IEngine engine = new Engine { Environment = { CurrentOutputStream = TextStream.Create(stringBuilder) } })
            {
                Consult(engine.Program);

                Term goal = Parse(@"twice(_), once(!), write('Forwards '), fail. ");
                Test(engine, goal, Failure(), "Expected failure.");
                Assert.AreEqual("C Forwards Moss Forwards ", stringBuilder.ToString());
            }
        }

        [TestMethod]
        public void Test_7_8_4_4_Examples_12()
        {
            var stringBuilder = new StringBuilder();
            using (IEngine engine = new Engine { Environment = { CurrentOutputStream = TextStream.Create(stringBuilder) } })
            {
                Consult(engine.Program);

                Term goal = Parse(@"twice(_), call(!), write('Forwards '), fail. ");
                Test(engine, goal, Failure(), "Expected failure.");
                Assert.AreEqual("C Forwards Moss Forwards ", stringBuilder.ToString());
            }
        }

        // ReSharper restore InconsistentNaming
        #endregion
    }
}
