﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Globalization;
using System.Reflection;
using System.Security;
using System.Windows.Forms;
using JetBrains.Annotations;
using OpcLabs.BaseLib.Browsing;
using OpcLabs.BaseLib.Collections.Specialized;
using OpcLabs.BaseLib.Facets;
using OpcLabs.BaseLib.Forms;
using System.Linq;
using System.Collections;
using System.Runtime.InteropServices;
using OpcLabs.BaseLib.Diagnostics;
using OpcLabs.BaseLib.Forms.Browsing;
using OpcLabs.BaseLib.Forms.Browsing.Parts;
using OpcLabs.BaseLib.Forms.Facets;
using OpcLabs.BaseLib.Forms.Extensions;

namespace OpcLabs.BaseLibFormsTestApp
{
    public partial class BrowseNodesTestSample : Form
    {
        #region Constructors

        public BrowseNodesTestSample()
        {
            InitializeComponent();
            propertyGrid1.ShowPropertyNameInHelp();
            propertyGrid1.SelectedObject = _formOptions;
        }

        #endregion

        #region Private classes

        static public class ObjectHelpers
        {
            static public object GetProperty(object o, string propertyName)
            {
                if (o == null) return null;
                if (String.IsNullOrEmpty(propertyName)) return o.ToString();

                Type objectType = o.GetType();
                PropertyInfo property = objectType.GetProperty(propertyName);
                return property.GetValue(o, null);
            }

            static public void SetProperty(object o, string propertyName, object value)
            {
                if (o == null) return;
                if (String.IsNullOrEmpty(propertyName)) return;

                Type objectType = o.GetType();
                PropertyInfo property = objectType.GetProperty(propertyName);
                property.SetValue(o, value, null);
            }
        }

        #region ZZ test
        private class BrowsePresentation : IBrowsePresentation
        {
            public IEnumerable<Guid> GetStaticColumns()
            {
                Check.EnsuresResultNotNull<IEnumerable<Guid>>();
                return Enumerable.Empty<Guid>();
            }

            public string GetText(BrowseNodesTextId id)
            {
                switch (id)
                {
                    case BrowseNodesTextId.Title: return "Property Dialog";
                    case BrowseNodesTextId.NodesListViewLabelFormat: return "Current node: {0} \nContent:";
                }
                return null;
            }

            public IEnumerable<ToolStripItem> GetToolMenuItems(IBrowsePresentationHost browsePresentationHost, BrowseNodesFormViews view)
            {
                Always.RequiresNotNull(browsePresentationHost, nameof(browsePresentationHost));
                Check.EnsuresResultNotNull<IEnumerable<ToolStripItem>>();
                return Enumerable.Empty<ToolStripItem>();
            }


            public IEnumerable<BrowseStateIcon> GetDefaultIcons()
            {
                Check.EnsuresResultNotNull<IEnumerable<BrowseStateIcon>>();
                Check.Ensures(Check.ForAll(Check.Result<IEnumerable<BrowseStateIcon>>(), browseIcon => browseIcon != null));

                return DefaultBrowsePresentation.Instance.GetDefaultIcons();
            }

            public BrowseColumnParameters GetDisplayNameColumnParameters(BrowseNodesFormViews view)
            {
                return DefaultBrowsePresentation.Instance.GetDisplayNameColumnParameters(view);
            }

            public string GetExceptionText(Exception exception)
            {
                Always.RequiresNotNull(exception, nameof(exception));
                Check.EnsuresResultNotNull<string>();

                return DefaultBrowsePresentation.Instance.GetExceptionText(exception);
            }

            public BrowseIcon GetIcon(BrowseNodesFormIcon iconId)
            {
                return DefaultBrowsePresentation.Instance.GetIcon(iconId);
            }

            public IEnumerable<Rectangle> GetIconLayout(BrowseNodesFormView view, BrowseImageSize imageSize)
            {
                Check.EnsuresResultNotNull<IEnumerable<Rectangle>>();
                return Enumerable.Empty<Rectangle>();
            }

            /*
            // from ...
            public BrowseNodesForm.FormOptions GetFormOptions()
            {
                var formOptions = new BrowseNodesForm.FormOptions
                {
                    //ShowListWithBranches = true,
                    ShowListWithLeaves = true,
                    ListAllowViewDetails = true,
                    ListAllowViewTile = true,
                    ListDefaultView = View.Details,
                    ShowRoot = true,

                    ShowImagesInNodesList = false,
                };
                return formOptions;
            }
             * */

            //from OpcBrowsePresentation

            public BrowseNodesPresentationOptions GetPresentationOptions()
        {
            Check.EnsuresResultNotNull<BrowseNodesPresentationOptions>();
            return new BrowseNodesPresentationOptions();
        }

        }

        private class NodePresentation : INodePresentation
        {
            public bool CreateNode(object parentNode, Type type, out object nodeCreated)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Always.RequiresNotNull(type, nameof(type));
                Check.Ensures(Check.Result<bool>() || (Check.ValueAtReturn(out nodeCreated) == null));

                nodeCreated = null;
                return false;
            }

            public bool EditNode(IEnumerable<object> oldNodePath, IBrowsePresentationHost browsePresentationHost, out object newNode)
            {
                Always.RequiresNotNull(oldNodePath, nameof(oldNodePath));
                IEnumerable<object> enumeratedNodePath = oldNodePath as IList<object> ?? oldNodePath.ToList();
                Always.Requires(Check.ForAll(enumeratedNodePath, o => o != null));
                Always.RequiresNotNull(browsePresentationHost, nameof(browsePresentationHost));
                Check.Ensures(Check.Result<bool>() || (Check.ValueAtReturn(out newNode) == null));

                newNode = null;
                return false;
            }

            public IEnumerable<BrowseStateIcon> GetBrowseStateIcons(Guid iconGuid)
            {
                Check.Ensures(Check.Result<IEnumerable<BrowseStateIcon>>() != null);
                // ReSharper disable once AssignNullToNotNullAttribute
                Check.Ensures(Check.ForAll(Check.Result<IEnumerable<BrowseStateIcon>>(), browseIcon => browseIcon != null));

                throw new ArgumentException("UnknownIconGuid", nameof(iconGuid));
            }

            public BrowseColumnParameters GetColumnParameters(Guid columnGuid)
            {
                if (columnGuid == PropertyNameColumnGuid)
                    // ReSharper disable LocalizableElement
                    return new BrowseColumnParameters("Name");
                // ReSharper restore LocalizableElement
                if (columnGuid == DataTypeColumnGuid)
                    // ReSharper disable LocalizableElement
                    return new BrowseColumnParameters("Data Type");
                // ReSharper restore LocalizableElement
                if (columnGuid == ItemIdColumnGuid)
                    // ReSharper disable LocalizableElement
                    return new BrowseColumnParameters("Item ID");
                // ReSharper restore LocalizableElement
                if (columnGuid == DescriptionColumnGuid)
                    // ReSharper disable LocalizableElement
                    return new BrowseColumnParameters("Description");
                // ReSharper restore LocalizableElement

                return DefaultNodePresentation.Instance.GetColumnParameters(columnGuid);
            }

            public IEnumerable<Guid> GetColumnsForParent(Object parentNode)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Check.EnsuresResultNotNull<IEnumerable<Guid>>();

                return new List<Guid> { PropertyNameColumnGuid, DataTypeColumnGuid, ItemIdColumnGuid, DescriptionColumnGuid };
            }


            public IEnumerable<Guid> GetTypeColumns(Type type)
            {
                Always.RequiresNotNull(type, nameof(type));
                Check.Ensures(Check.Result<IEnumerable<Guid>>() != null);

                return new List<Guid> { PropertyNameColumnGuid, DataTypeColumnGuid, ItemIdColumnGuid, DescriptionColumnGuid };
            }

            public IEnumerable<Guid> GetTypeIconGuids(Type type)
            {
                Always.RequiresNotNull(type, nameof(type));
                Check.EnsuresResultNotNull<IEnumerable<Guid>>();

                return DefaultNodePresentation.Instance.GetTypeIconGuids(type);
            }

            public string GetContent(object node, Guid columnGuid)
            {
                Always.RequiresNotNull(node, nameof(node));
                return DefaultNodePresentation.Instance.GetContent(node, columnGuid);
            }

            public IEnumerable<ToolStripItem> GetContextMenuItems(IEnumerable<object> nodePath, bool isManual, BrowseNodesFormViews view, IBrowsePresentationHost browsePresentationHost)
            {
                // ReSharper disable once PossibleMultipleEnumeration
                Always.RequiresAllNotNull(nodePath, nameof(nodePath));
                Always.RequiresNotNull(browsePresentationHost, nameof(browsePresentationHost));
                Check.EnsuresResultNotNull<IEnumerable<ToolStripItem>>();

                var items = new List<ToolStripItem>();

                if (isManual)
                {
                    var menuItem = new ToolStripMenuItem("Delete Manual") {ShortcutKeys = Keys.F8};
                    menuItem.Click += (sender, e) => { browsePresentationHost.DeleteManualNode(nodePath); };
                    items.Add(menuItem);

                    menuItem = new ToolStripMenuItem("Is manual");
                    menuItem.Click += (sender, e) => { MessageBox.Show("Is manual: " + browsePresentationHost.IsNodeManual(nodePath)); };
                    items.Add(menuItem);
                }
                return items;
            }

            public string GetCreateNodeNameHint(object parentNode, Type type)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Always.RequiresNotNull(type, nameof(type));

                return DefaultNodePresentation.Instance.GetCreateNodeNameHint(parentNode, type);
            }

            public bool GetDataObject(IEnumerable<IEnumerable<object>> nodePaths, BrowseNodesFormViews view, out IDataObject dataObject)
            {
                Always.RequiresNotNull(nodePaths, nameof(nodePaths));

                return DefaultNodePresentation.Instance.GetDataObject(nodePaths, view, out dataObject);
            }


            public object GetDefaultBranchFilter(object parentNode)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                return DefaultNodePresentation.Instance.GetDefaultBranchFilter(parentNode);
            }


            public object GetDefaultLeafFilter(object parentNode)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                return DefaultNodePresentation.Instance.GetDefaultLeafFilter(parentNode);
            }

            public string GetDisplayName(IEnumerable<object> nodePath, BrowseNodesFormViews view)
            {
                Always.RequiresNotNull(nodePath, nameof(nodePath));
                IEnumerable<object> enumeratedNodePath = nodePath as IList<object> ?? nodePath.ToList();
                Always.Requires(Check.ForAll(enumeratedNodePath, o => o != null));
                return DefaultNodePresentation.Instance.GetDisplayName(enumeratedNodePath, view);
            }

            public bool GetIconGuidLegend(Guid iconGuid, out string category, out string displayName, out string description)
            {
                return DefaultNodePresentation.Instance.GetIconGuidLegend(iconGuid, out category, out displayName, out description);
            }

            public bool GetTextForEditing(IEnumerable<object> parentNodePath, object node, Type type,
                BrowseNodesFormViews view, ref string text, ref int selectionStart, ref int selectionLength, ref Font font)
            {
                Always.RequiresNotNull(parentNodePath, nameof(parentNodePath));
                IEnumerable<object> enumeratedParentNodePath = parentNodePath as IList<object> ?? parentNodePath.ToList();
                Always.Requires(Check.ForAll(enumeratedParentNodePath, o => o != null));
                Always.RequiresNotNull(type, nameof(type));
                return DefaultNodePresentation.Instance.GetTextForEditing(enumeratedParentNodePath, node, type, view, ref text, ref selectionStart, ref selectionLength, ref font);
            }

            public string GetTypeName(Type type)
            {
                Always.RequiresNotNull(type, nameof(type));
                return DefaultNodePresentation.Instance.GetTypeName(type);
            }

           public string GetInfoText(Object node)
            {
                Always.RequiresNotNull(node, nameof(node));
                return DefaultNodePresentation.Instance.GetInfoText(node);
            }

           public IEnumerable<Guid?> GetNodeIconGuids(object node)
           {
               Always.RequiresNotNull(node, nameof(node));
               Check.EnsuresResultNotNull<IEnumerable<Guid?>>();

               return Enumerable.Empty<Guid?>();
            }

            public bool ModifyNodeStyle(object node, BrowseIconStates nodeStates, BrowseNodeStyle nodeStyle)
            {
                Always.RequiresNotNull(node, nameof(node));
                Always.RequiresNotNull(nodeStyle, nameof(nodeStyle));

                return DefaultNodePresentation.Instance.ModifyNodeStyle(node, nodeStates, nodeStyle);
            }

            public void NodeCreated(object node)
            {
                Always.RequiresNotNull(node, nameof(node));
                // Do nothing.
            }

            public void NodeEdited(object node, object oldNode)
            {
                Always.RequiresNotNull(node, nameof(node));
                Always.RequiresNotNull(oldNode, nameof(oldNode));
                // Do nothing.
            }

            [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1062:ValidateArgumentsOfPublicMethods")]    // using CC runtime check on interface
            public bool ShouldAutoExpand(IEnumerable<object> nodePath)
            {
                Always.RequiresNotNull(nodePath, nameof(nodePath));
                IEnumerable<object> enumeratedNodePath = nodePath as IList<object> ?? nodePath.ToList();
                Always.Requires(Check.ForAll(enumeratedNodePath, o => o != null));

                return DefaultNodePresentation.Instance.ShouldAutoExpand(enumeratedNodePath);
            }

        }

        private class Browseable : IBrowseable
        {
            public void BrowseBranches(object parentNode, object nodeFilter, out IEnumerable<object> branches, out IEnumerable<Exception> exceptions)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Check.Ensures(Check.ValueAtReturn(out branches) != null);
                Check.Ensures(Check.ValueAtReturn(out exceptions) != null);

                branches = Enumerable.Empty<Object>();
                exceptions = Enumerable.Empty<Exception>();
            }

            public void BrowseLeaves(object parentNode, object nodeFilter, out IEnumerable<object> leaves, out IEnumerable<Exception> exceptions)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Check.Ensures(Check.ValueAtReturn(out leaves) != null);
                Check.Ensures(Check.ValueAtReturn(out exceptions) != null);

                leaves = new object[] { "o1", "o2" };
                exceptions = Enumerable.Empty<Exception>();
            }

            public void CreateNode(object parentNode, Type type, string name, out object node, out Exception exception)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Always.RequiresNotNull(type, nameof(type));
                Always.RequiresNotNull(name, nameof(name));
                Check.Ensures(Check.ValueAtReturn(out node) != null || Check.ValueAtReturn(out exception) != null);

                throw new NotImplementedException();
            }

            public IEnumerable<Type> GetAddableBranchTypes(object parentNode, bool onlyUserAddable)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Check.EnsuresResultNotNull<IEnumerable<Type>>();

                return Enumerable.Empty<Type>();
            }

            public IEnumerable<Type> GetAddableLeafTypes(object parentNode, bool onlyUserAddable)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Check.EnsuresResultNotNull<IEnumerable<Type>>();

                return Enumerable.Empty<Type>();
            }

            public void HasChildren(object parentNode, out bool parentHasBranches, out bool parentHasLeaves, out IEnumerable<Exception> exceptions)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Check.Ensures(Check.ValueAtReturn(out exceptions) != null);

                parentHasBranches = true;
                parentHasLeaves = true;
                exceptions = Enumerable.Empty<Exception>();
            }

            public void InvalidateCache(object parentNode)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                throw new NotImplementedException();
            }

            public bool IsBranch(object parentNode, object node)
            {
                Always.RequiresNotNull(node, nameof(node));
                return true;
            }

            public bool IsLeaf(object parentNode, object node)
            {
                Always.RequiresNotNull(node, nameof(node));
                return true;
            }

            public bool IsMyNode(object parentNode, object node)
            {
                Always.RequiresNotNull(node, nameof(node));
                return true;
            }

            public bool IsSelectable(object parentNode, object node)
            {
                Always.RequiresNotNull(node, nameof(node));
                return true;
            }

            #region IBrowseable Members


            public bool MatchNode(object parentNode, object nodeA, object nodeB)
            {
                Always.RequiresNotNull(nodeA, nameof(nodeA));
                Always.RequiresNotNull(nodeB, nameof(nodeB));

                return Equals(nodeB, nodeA);
            }

            public void NodeAdded(object parentNode, object node)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Always.RequiresNotNull(node, nameof(node));
                // do nothing
            }

            #endregion
        }

        private class NodeIdentification : INodeIdentification
        {
            public string GetNodeId(Object node)
            {
                Always.RequiresNotNull(node, nameof(node));
                Check.EnsuresResultNotNull<string>();

                string s = node.ToString();
                if (s == "bbb") return "";
                return "id_" + s;
            }

            public bool ValidateNodeId(string nodeId, out string errorMessage)
            {
                Always.RequiresNotNull(nodeId, nameof(nodeId));

                errorMessage = null;
                if (nodeId.Length > 2) return true;
                errorMessage = "Id is short.";
                return false;
            }

        }
        #endregion

        #endregion

        #region Events

        private readonly BrowseNodesControlOptions _formOptions = new BrowseNodesControlOptions
        {
            ShowAddNewBranches = true,
            ShowAddNewLeaves = true,
            ShowDeleteButton = true,
            ShowToolsButton = true,
            ShowBranchesFilter = true,
            ShowLeavesFilter = true,
            ShowTreeWithBranches = true,
            ShowListWithBranches = true,
            ShowListWithLeaves = true,
            ListAllowAllViews = true,
            ShowSelectedNodesList = true,
            SelectedListAllowAllViews = true,

            ShowFormInformationLabel = true,
            ShowFormInformationBox = true,

            ShowNodeInformationLabel = true,
            ShowNodeInformationBox = true,
            NodeInformationBoxHeight = 130,

            ShowNodeIdentificationLabel = true,
            ShowNodeIdentificationBox = true,

            SearchBoxOptions = new BaseLib.Forms.Common.SearchBoxOptions(),
            ShowSearchBox = true,

            ShowActionsPane = true,
            ActionsPaneVariableWidthAllowed = true,
            ShowLegendPane = true,
            LegendPaneVariableHeightAllowed = true,
        };

        private Dictionary<Object, Object> _branchFilters = null;
        private Dictionary<Object, Object> _leafFilters = null; 

        // ReSharper disable InconsistentNaming
        private void button1_Click(object sender, EventArgs e)
// ReSharper restore InconsistentNaming
        {
            ShowBrowseNodes(false);
        }

        private OpcDANodeDialog form = null;

        private void ShowBrowseNodes(bool asControl)
        {
//            OpcLabs.EasyOpc.DataAccess.Forms.BrowseObjectsForm form = new OpcLabs.EasyOpc.DataAccess.Forms.BrowseObjectsForm();
            var browsePresentation = new BrowsePresentationTestSample();
            var nodePresentation = new NodePresentationTestSample();

            /*
// ReSharper disable LoopCanBeConvertedToQuery
            foreach (GroupBox groupBox in new[] { showControlsGroupBox, listViewGroupBox, selectedListViewGroupBox })
// ReSharper restore LoopCanBeConvertedToQuery
            {
// ReSharper disable LoopCanBeConvertedToQuery
                foreach (Control control in groupBox.Controls)
// ReSharper restore LoopCanBeConvertedToQuery
                {
                    var checkBox = control as CheckBox;
                    if ((checkBox != null) && (checkBox.Checked) && (checkBox.Tag is string))
                    {
                        ObjectHelpers.SetProperty(formOptions, (string)control.Tag, true);
                    }
                }
            }
            formOptions.FormWidth = 300;
            formOptions.ShowFocusedNodeInformationLabel = true;
            formOptions.ShowFocusedNodeInformationText = true;
             */
            //browsePresentation.SetFormOptions(_formOptions);


// ReSharper disable UseObjectOrCollectionInitializer

        //    var form = new OpcDANodeDialog(asControl);
            if (form == null) form = new OpcDANodeDialog(asControl);
            
// ReSharper restore UseObjectOrCollectionInitializer

            int test = 0;

            form.Mode = selectableRadioButton.Checked ? BrowseNodesFormMode.Selectable : BrowseNodesFormMode.NodeIdentification;
            form.MultiSelect = multiSelectCheckBox.Checked;
            switch (test)
            {
                case 0:
                    form.Browseable = new BrowseableTestSample();

                    form.NodeIdentification = new NodeIdentification();
                    form.BrowsePresentation = browsePresentation;
                    form.Options = _formOptions;
                    form.NodePresentation = nodePresentation;
                    form.SelectableFilter = Test;
                    form.TerminalNodePlace = BrowseTerminalNodePlace.PreferList;
                    //                    form.CurrentPath = new Object[] { new Root("RootForm1"), new Network("Second"), new Folder("5"), new Folder("L1") };
                    form.CurrentPath = new Object[] { new Root("RootForm1"), new Network("Second"), new Folder("3") };
                    //form.CurrentPath = new Object[] { new Root("RootForm1") };
                    //                    form.CurrentPath = new Object[] { new Root("RootForm1"), new Folder("XXX") };

#pragma warning disable 168
                    var list = new List<IEnumerable<Object>>
#pragma warning restore 168
                           {
//                               new Object[] { new Root("RootForm1"), new Network("Second"), new Folder("no5") },
//////                               new Object[] { new Root("RootForm1"), new Network("Second"), new Folder("5") },
                               new Object[] { new Root("RootForm1"), new Network("Second"), new Folder("6") },
//                               new Object[] { new Root("RootForm1"), new Network("Second"), new Folder("2") },
//                               new object[] { new Root("RootForm1"), new Network("Three"), new WorkGroup("Three Group")},
                               //new object[] { new Root("RootForm1"), new Network("Second")},
            //                   new object[] { new Root("RootForm1"), new Network("ABCD"), new Folder("2"), "L5"},
              //                 new object[] { new Root("RootForm1"), new Folder("Fold"), new Folder("2"), new Folder("FF"), "kk"},
                //               new object[] { new Root("RootForm1"), new Network("Second"), new Folder("5"), "L0"},
                  //             new object[] { new Root("RootForm1"), new Network("Second"), new Folder("5"), "A"},
                    //           new object[] { new Root("RootForm1"), new Network("Second"), new Folder("5"), new WorkGroup("WW")},
                           };
            //form.SelectedNodes = list;
            form.SelectedNodePaths = list;

            var manuallyAddedNodePaths = new List<IEnumerable<Object>>
            {
                new object[] {new Root("RootForm1"), new Network("Second"), new Folder("2"), new WorkGroup("W5") },
                new object[] {new Root("RootForm1"), new Network("Second"), new Folder("2"), new Folder("F5") },
                new object[] {new Root("RootForm1"), new Network("Second"), new Folder("6"), new Folder("F5") },
                new object[] {new Root("RootForm1"), new Network("Second"), new Folder("2"), new WorkGroup("Leaf1") },
            };
                    //form.ManuallyAddedNodePaths = manuallyAddedNodePaths;
              var manuallyAddedLeaves = new List<IEnumerable<Object>>
            {
                new object[] {new Root("RootForm1"), new Network("Second"), new Folder("2"), new WorkGroup("Leaf1") },
            };

//                    form.ManuallyAddedLeaves = manuallyAddedLeaves;
              form.MinimumLevel = 1;
                    break;

                case 1:
                    form.Browseable = new BrowseableNetwork();

                    form.NodeIdentification = new NodeIdentification();
                    form.BrowsePresentation = browsePresentation;
                    form.Options = _formOptions;
                    form.NodePresentation = nodePresentation;
                    form.CurrentPath = new Object[] { new BrowseableNetwork.RootNetworkNode("Root") };
                    form.MinimumLevel = 1;
                    break;

                case 2:



                    var formOptions = new BrowseNodesControlOptions
                    {
                        ShowFormInformationBox = true,
                        ShowListWithLeaves = true,
                        ShowUpButton = false,
                        ListAllowViewDetails = true,
                        //ListDefaultView = View.Details,
                        //ShowRoot = true,
                        ShowImagesInNodesList = true,
                        ShowImagesInNodesTree = false,
                        ShowImagesInSelectedNodesList = false,
                        ShowNodeInformationLabel = true,
                        ShowNodeInformationBox = true,
                        ToolBarLocation = BrowseNodesToolBarLocation.RightOnFormInformationPanel,
                        ShowStatusInformation = true,

                        //FormWidth = 465,
                        //FormHeight = 448,
                        //FormMinimumWidth = 305,
                        //FormMinimumHeight = 375,
                        ToolBarVerticalHeight = 1 * 23 + 16,
                        FormInformationBoxHeight = 2 * 13 + 3,
                        NodeInformationBoxHeight = 6 * 13 + 3,  // 1 line is spare for possible wrapping
                    };

                    
                    var currentPath = new List<object> { new Root("RootForm1") };

                //form.Browseable = new ServerBrowseable(ServerFamilies, easyDAClient1, easyAEClient1);
                    form.Browseable = new BrowseableTestSample();
                    form.MultiSelect = false;
                    form.CurrentPath = currentPath;
                    form.MinimumLevel = 1;
                //browsePresentation.SetFormOptions(formOptions)
                form.Options = formOptions;

                form.BrowsePresentation = browsePresentation;

                form.NodePresentation = nodePresentation;
                //form.NodePresentation = new CompositeNodePresentation
                //                                       {new ServerNodePresentation(), new ServerNodePresentation()};


                     

                break;
        }
            //form.Show();
            //try
            //{
            if (_branchFilters != null) form.BranchFilters = _branchFilters;
            if (_leafFilters != null) form.LeafFilters = _leafFilters;
            //form.SelectedNodePaths = Enumerable.Empty<IEnumerable<object>>();
            if (showInspectorPanelCheckBox.Checked)
            {
                //IFacetProvider<InspectorFacet>
                form.InspectorFacetProvider = new TestFacetProvider();
            }
            
            form.ShowDialog();
            var outCurrentPath = form.CurrentPath;
            _branchFilters = form.BranchFilters;
            _leafFilters = form.LeafFilters;
            //}
            
            //catch (Exception exc)
            //{
// ReSharper disable LocalizableElement
            //    MessageBox.Show(exc.Message, Text + " - Error Message", MessageBoxButtons.OK, MessageBoxIcon.Error);
// ReSharper restore LocalizableElement
            //}
            



        }


        // ReSharper disable InconsistentNaming
        private void button2_Click(object sender, EventArgs e)
        // ReSharper restore InconsistentNaming
        {
            // ReSharper disable UseObjectOrCollectionInitializer
#pragma warning disable IDE0017 // Simplify object initialization
            var browseNodesForm = new BrowseNodesForm();
#pragma warning restore IDE0017 // Simplify object initialization
                               // ReSharper restore UseObjectOrCollectionInitializer

            browseNodesForm.Browseable = new Browseable
            // ReSharper disable RedundantEmptyObjectOrCollectionInitializer
            {
                //EasyDAClient = easyDAClient1,
                //ServerDescriptor = ServerDescriptor,
                //NodeDescriptor = NodeDescriptor
            };
            // ReSharper restore RedundantEmptyObjectOrCollectionInitializer
            var currentPath = new List<object> { new object() };
            browseNodesForm.CurrentPath = currentPath;
            browseNodesForm.MinimumLevel = 1;
            browseNodesForm.BrowsePresentation = new BrowsePresentation();
            browseNodesForm.Options = new BrowseNodesControlOptions
            {
                //ShowFormInformationBox = true,
                ShowListWithLeaves = true,
                ShowListWithBranches = true,
                ShowTreeWithBranches = true,
                ShowSelectedNodesList = true,
                ListAllowViewDetails = true,
                //ListDefaultView = View.Details,
                //ShowRoot = true,
                ShowImagesInNodesList = true,
                ShowImagesInNodesTree = true,
                ShowImagesInSelectedNodesList = true,
                ToolBarLocation = BrowseNodesToolBarLocation.TopOnForm,
                ShowNodeInformationBox = true,

                //FormWidth = 800,
                //FormHeight = 600,
                //FormMinimumWidth = 480,
                //FormMinimumHeight = 320,
                TreeBoxWidth = -35, // -percent
                ListBoxWidth = -25, // -percent
                SelectedBoxWidth = -40, // -percent
                TreeBoxMinimumWidth = 70,
                ListBoxMinimumWidth = 50,
                SelectedListBoxMinimumWidth = 80,
                //ToolBarVerticalHeight = 2 * 23 + 16,
                //FormInformationBoxHeight = 2 * 13 + 5,
                NodeInformationBoxHeight = 5 * 13 + 5,
            };

            browseNodesForm.NodePresentation = new NodePresentation();

            // ReSharper disable RedundantJumpStatement
            if (browseNodesForm.ShowDialog() != DialogResult.OK) return;
            // ReSharper restore RedundantJumpStatement

        }


        #endregion

        #region Private Methods

// ReSharper disable UnusedMember.Local
        private bool Test(Object o)
// ReSharper restore UnusedMember.Local
        {
            if (o is string) return false;
            if (o.ToString().StartsWith("no", StringComparison.CurrentCultureIgnoreCase)) return false;
            return true;
        }

        #endregion

        #region Private Fields

        // temporary ZZ test
        static private readonly Guid PropertyNameColumnGuid = new Guid("{9BCD790C-97A4-4246-AC80-185F023D8E79}");
        static private readonly Guid DataTypeColumnGuid = new Guid("{6C0BAD97-15F7-4FB6-8189-1BDF0E1C6995}");
        static private readonly Guid ItemIdColumnGuid = new Guid("{4C9B70EE-D02B-4B49-A834-ABF2E1ABCF44}");
        static private readonly Guid DescriptionColumnGuid = new Guid("{78358F04-D33B-4550-992B-51C453B19F15}");

        #endregion

        public class TestInspectorFacet : InspectorFacet
        {
            public TestInspectorFacet([NotNull] string title, [CanBeNull] Image image, [CanBeNull] string description)
                : base(title, image)
            {
                _description = description;
            }

            public override Control CreateControl()
            {
                Check.EnsuresResultNotNull<Control>();

                if (_description == null) return null;
                var panel = new Panel();
                //var richTextBox = new RichTextBox();
                //richTextBox.UpdateText(_description);
                //richTextBox.ContextMenuStrip = new RichTextBoxContextMenuStrip();
                //panel.Controls.Add(richTextBox);
                //richTextBox.Dock = DockStyle.Top;
                var textBox = new TextBox {Text = _description};
                panel.Controls.Add(textBox);
                textBox.Dock = DockStyle.Top;
                panel.Dock = DockStyle.Fill;
                panel.BackColor = SystemColors.Control;
                return panel;
            }

            [CanBeNull]
            private readonly string _description;
        }


        public class TestFacetProvider : IFacetProvider<InspectorFacet>
        {
            static private readonly Guid TestFacetGuid0 = new Guid("{BEFADF70-9E43-469E-978F-B42381801CEA}");
            static private readonly Guid TestFacetGuid1 = new Guid("{DCD60CF3-E0EC-4051-881F-692A997412D3}");
            static private readonly Guid TestFacetGuid2 = new Guid("{5D19D393-4382-457A-ACC7-14D5ADC9EAFC}");
            static private readonly Guid TestFacetGuid3 = new Guid("{E9C54623-8260-43C6-B402-01692B34B0AB}");

            public ISet<Guid> GetFacetGuids(object obj)
            {
                Always.RequiresNotNull(obj, nameof(obj));
                Check.EnsuresResultNotNull<ISet<Guid>>();

                switch (obj.GetType().Name)
                {
                    case "Folder": return new GuidSet(new[] { TestFacetGuid0, TestFacetGuid1 });
                    case "WorkGroup": return new GuidSet();
                    default: return new GuidSet(new[] { TestFacetGuid0 });
                }
            }

            public InspectorFacet GetObjectFacet(object obj, Guid facetGuid)
            {
                Always.RequiresNotNull(obj, nameof(obj));

                if (facetGuid == TestFacetGuid0) return new TestInspectorFacet("Description", null, obj.ToString());
                if (facetGuid == TestFacetGuid1) return new TestInspectorFacet("Folder", null, obj.ToString());
                return null;
            }
        }

        /// <summary>
        /// Displays a dialog box from which the user can select an OPC node. 
        /// </summary>
        /// <remarks>
        /// <para>OpcDANodeDialog is a modal dialog box; therefore, when shown, it blocks the rest of the application until the 
        /// user has chosen an OPC item. When a dialog box is displayed modally, no input (keyboard or mouse click) can occur except 
        /// to objects on the dialog box. The program must hide or close the dialog box (usually in response to some user action) 
        /// before input to the calling program can occur. </para>
        /// </remarks>
        // ReSharper disable InconsistentNaming
        public partial class OpcDANodeDialog : ConcreteCommonDialog
        // ReSharper restore InconsistentNaming
        {
            /// <summary>
            /// Initializes an instance of the <see cref="OpcDANodeDialog"/> class. 
            /// </summary>
            public OpcDANodeDialog(bool asControl)
            {
                if (asControl)
                {
                    var f1 = new BrowseNodesControlTestForm();
                    _browseNodesForm = f1;
                    _browseNodesControl = f1.browseNodesControl3;
                }
                else
                {
                    _browseNodesForm = new BrowseNodesForm();
                    _browseNodesControl = _browseNodesForm as IBrowseNodesControl;
                }
                InitializeComponent();
            }

            /// <summary>
            /// Initializes an instance of the <see cref="OpcDANodeDialog"/> class with Designer support. 
            /// </summary>
            /// <param name="container">The component container to which the new instance will be added.</param>
            public OpcDANodeDialog(IContainer container)
            {
                if (container == null)
                {
                    throw new ArgumentNullException(nameof(container));
                }
                container.Add(this);

                InitializeComponent();
            }

            [CanBeNull]
            public IFacetProvider<InspectorFacet> InspectorFacetProvider
            {
                get
                {
                    if (_browseNodesForm is BrowseNodesForm browseNodesForm) return browseNodesForm.InspectorFacetProvider;
                    return null;
                } 
                set
                {
                    if (_browseNodesForm is BrowseNodesForm browseNodesForm)
                    {
                        browseNodesForm.InspectorFacetProvider = value;
                        browseNodesForm.InspectorPanelSize = new Size(200, 200);
                        browseNodesForm.ShowInspectorPanel = true;
                    }
                }
            } 

            internal IBrowseNodesControl SecondControl
            {
                get
                {
                    var browseNodesControlTestForm = _browseNodesForm as BrowseNodesControlTestForm;
                    return browseNodesControlTestForm?.browseNodesControl4;
                }
            }

            private bool SetSecondControl => true && (SecondControl != null);

            public BrowseNodeKinds AllowedSelectedKinds { get => _browseNodesControl.AllowedSelectedKinds;
                set => _browseNodesControl.AllowedSelectedKinds = value;
            }

            /// <summary>
            /// Sets Browseable interface.
            /// </summary>
// ReSharper disable MemberHidesStaticFromOuterClass
            public IBrowseable Browseable
// ReSharper restore MemberHidesStaticFromOuterClass
            {
                get => _browseNodesControl.Browseable;
                set { _browseNodesControl.Browseable = value;
                    if (SetSecondControl) SecondControl.Browseable = value; 
                }
            }

            /// <summary>
            /// Sets BrowsePresenation interface.
            /// </summary>
            // ReSharper disable MemberHidesStaticFromOuterClass
            public IBrowsePresentation BrowsePresentation
            // ReSharper restore MemberHidesStaticFromOuterClass
            {
                get => _browseNodesControl.BrowsePresentation;
                set { _browseNodesControl.BrowsePresentation = value; if (SetSecondControl) SecondControl.BrowsePresentation = value; }
            }

            /// <summary>
            /// Sets BrowsePresenation interface.
            /// </summary>
            // ReSharper disable MemberHidesStaticFromOuterClass
            public INodePresentation NodePresentation
            // ReSharper restore MemberHidesStaticFromOuterClass
            {
                get => _browseNodesControl.NodePresentation;
                set { _browseNodesControl.NodePresentation = value; if (SetSecondControl) SecondControl.NodePresentation = value; }
            }

            public BrowseNodesControlOptions Options
            {
                get => _browseNodesControl.Options;
                set { _browseNodesControl.Options = value; if (SetSecondControl) SecondControl.Options = value; }
                
            }

            /// <summary>
            /// Gets or sets current path.
            /// </summary>
            public IEnumerable<Object> CurrentPath
            {
                get => _browseNodesControl.CurrentPath;
                set { _browseNodesControl.CurrentPath = value; if (SetSecondControl) SecondControl.CurrentPath = value; }
            }

            /// <summary>
            /// Gets or sets set of manually added node paths.
            /// </summary>
            public IEnumerable<IEnumerable<Object>> ManuallyAddedNodePaths
            {
                get => _browseNodesControl.ManuallyAddedNodePaths;
                set { _browseNodesControl.ManuallyAddedNodePaths = value; if (SetSecondControl) SecondControl.ManuallyAddedNodePaths = value; }
            }

            /// <summary>
            /// Sets minimum level.
            /// </summary>
            public int MinimumLevel
            {
                get => _browseNodesControl.MinimumLevel;
                set { _browseNodesControl.MinimumLevel = value; if (SetSecondControl) SecondControl.MinimumLevel = value; }
            }

            public BrowseNodesFormMode Mode
            {
                get => _browseNodesControl.Mode;
                set { _browseNodesControl.Mode = value; if (SetSecondControl) SecondControl.Mode = value; }
            }

            /// <summary>
            /// Sets a value indicating whether multi nodes can be selected.
            /// </summary>
            public bool MultiSelect
            {
                get => _browseNodesControl.MultiSelect;
                set { _browseNodesControl.MultiSelect = value; if (SetSecondControl) SecondControl.MultiSelect = value; }
            }

            /// <summary>
            /// Gets pr sets NodeIdentification interface.
            /// </summary>
// ReSharper disable MemberHidesStaticFromOuterClass
            public INodeIdentification NodeIdentification
// ReSharper restore MemberHidesStaticFromOuterClass
            {
                get => _browseNodesControl.NodeIdentification;
                set { _browseNodesControl.NodeIdentification = value; if (SetSecondControl) SecondControl.NodeIdentification = value; }
            }


            /// <summary>
            /// Sets selectable filter.
            /// </summary>
            public Predicate<Object> SelectableFilter
            {
                get => _browseNodesControl.SelectableFilter;
                set { _browseNodesControl.SelectableFilter = value; if (SetSecondControl) SecondControl.SelectableFilter = value; }
            }

            public string SelectedNodeIdentification => _browseNodesControl.SelectedNodeId;

            /// <summary>
            /// Gets or sets set of selected node paths.
            /// </summary>
            public IEnumerable<IEnumerable<Object>> SelectedNodePaths
            {
                get => _browseNodesControl.SelectedNodePaths;
                set { _browseNodesControl.SelectedNodePaths = value; if (SetSecondControl) SecondControl.SelectedNodePaths = value; }
            }

            public BrowseTerminalNodePlace TerminalNodePlace
            {
                get => _browseNodesControl.TerminalNodePlace;
                set { _browseNodesControl.TerminalNodePlace = value; if (SetSecondControl) SecondControl.TerminalNodePlace = value; }
            }

            public Dictionary<Object, Object> BranchFilters
            {
                get => _browseNodesControl.BranchFilters;
                set { _browseNodesControl.BranchFilters = value; if (SetSecondControl) SecondControl.BranchFilters = value; }
            }

            public Dictionary<Object, Object> LeafFilters
            {
                get => _browseNodesControl.LeafFilters;
                set { _browseNodesControl.LeafFilters = value; if (SetSecondControl) SecondControl.LeafFilters = value; }
            }

            /// <summary>
            /// Specifies a common dialog box.
            /// </summary>
            /// <param name="hwndOwner">A value that represents the window handle of the owner window for the common dialog box.
            /// </param>
            /// <returns><b>true</b> if the dialog box was successfully run; otherwise, <b>false</b>.</returns>
            protected override bool RunDialog(IntPtr hwndOwner)
            {
                //_browseObjectsForm.ResetControls();
                if (_browseNodesForm.ShowDialog() != DialogResult.OK)
                    return false;
                return true;
            }

            [NotNull]
            private readonly Form _browseNodesForm; // = new BrowseNodesForm();
            [NotNull]
            private readonly IBrowseNodesControl _browseNodesControl; // = new BrowseNodesForm();
        }

// ReSharper disable InconsistentNaming
        partial class OpcDANodeDialog
// ReSharper restore InconsistentNaming
        {
            /// <summary>
            /// Required designer variable.
            /// </summary>
            private IContainer components;

            /// <summary> 
            /// Clean up any resources being used.
            /// </summary>
            /// <param name="disposing">true if managed resources should be disposed; otherwise, false.</param>
            protected override void Dispose(bool disposing)
            {
                if (disposing && (components != null))
                {
                    components.Dispose();
                }
                base.Dispose(disposing);
            }

            #region Component Designer generated code

            /// <summary>
            /// Required method for Designer support - do not modify
            /// the contents of this method with the code editor.
            /// </summary>
            private void InitializeComponent()
            {
                components = new System.ComponentModel.Container();
            }

            #endregion
        }


        #region Browse Network

        // For related or similar ideas or implementations, see also:
        // http://msdn.microsoft.com/en-us/library/aa370623%28VS.85%29.aspx

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1063:ImplementIDisposableCorrectly")] // temporary
        public class CompEnum : IEnumerable, IDisposable
        {
            #region "Server type enumeration"
            // Possible types of servers
            [Flags]
            public enum ServerType : uint
            {
                // ReSharper disable InconsistentNaming
                /// <summary>
                /// All workstations
                /// </summary>
                SV_TYPE_WORKSTATION = 0x00000001,
                /// <summary>
                /// All computers that have the server service running
                /// </summary>
                SV_TYPE_SERVER = 0x00000002,
                /// <summary>
                /// Any server running Microsoft SQL Server
                /// </summary>
                SV_TYPE_SQLSERVER = 0x00000004,
                /// <summary>
                /// Primary domain controller
                /// </summary>
                SV_TYPE_DOMAIN_CTRL = 0x00000008,
                /// <summary>
                /// Backup domain controller
                /// </summary>
                SV_TYPE_DOMAIN_BAKCTRL = 0x00000010,
                /// <summary>
                /// Server running the Timesource service
                /// </summary>
                SV_TYPE_TIME_SOURCE = 0x00000020,
                /// <summary>
                /// Apple File Protocol servers
                /// </summary>
                SV_TYPE_AFP = 0x00000040,
                /// <summary>
                /// Novell servers
                /// </summary>
                SV_TYPE_NOVELL = 0x00000080,
                /// <summary>
                /// LAN Manager 2.x domain member
                /// </summary>
                SV_TYPE_DOMAIN_MEMBER = 0x00000100,
                /// <summary>
                /// Server sharing print queue
                /// </summary>
                SV_TYPE_PRINTQ_SERVER = 0x00000200,
                /// <summary>
                /// Server running dial-in service
                /// </summary>
                SV_TYPE_DIALIN_SERVER = 0x00000400,
                /// <summary>
                /// Xenix server
                /// </summary>
                SV_TYPE_XENIX_SERVER = 0x00000800,
                /// <summary>
                /// Windows NT workstation or server
                /// </summary>
                SV_TYPE_NT = 0x00001000,
                /// <summary>
                /// Server running Windows for Workgroups
                /// </summary>
                SV_TYPE_WFW = 0x00002000,
                /// <summary>
                /// Microsoft File and Print for NetWare
                /// </summary>
                SV_TYPE_SERVER_MFPN = 0x00004000,
                /// <summary>
                /// Server that is not a domain controller
                /// </summary>
                SV_TYPE_SERVER_NT = 0x00008000,
                /// <summary>
                /// Server that can run the browser service
                /// </summary>
                SV_TYPE_POTENTIAL_BROWSER = 0x00010000,
                /// <summary>
                /// Server running a browser service as backup
                /// </summary>
                SV_TYPE_BACKUP_BROWSER = 0x00020000,
                /// <summary>
                /// Server running the master browser service
                /// </summary>
                SV_TYPE_MASTER_BROWSER = 0x00040000,
                /// <summary>
                /// Server running the domain master browser
                /// </summary>
                SV_TYPE_DOMAIN_MASTER = 0x00080000,
                /// <summary>
                /// Windows 95 or later
                /// </summary>
                SV_TYPE_WINDOWS = 0x00400000,
                /// <summary>
                /// Root of a DFS tree
                /// </summary>
                SV_TYPE_DFS = 0x00800000,
                /// <summary>
                /// Terminal Server
                /// </summary>
                SV_TYPE_TERMINALSERVER = 0x02000000,
                /// <summary>
                /// Server clusters available in the domain
                /// </summary>
                SV_TYPE_CLUSTER_NT = 0x01000000,
                /// <summary>
                /// Cluster virtual servers available in the domain
                /// (Not supported for Windows 2000/NT)
                /// </summary>			
                SV_TYPE_CLUSTER_VS_NT = 0x04000000,
                /// <summary>
                /// IBM DSS (Directory and Security Services) or equivalent
                /// </summary>
                SV_TYPE_DCE = 0x10000000,
                /// <summary>
                /// Return list for alternate transport
                /// </summary>
                SV_TYPE_ALTERNATE_XPORT = 0x20000000,
                /// <summary>
                /// Return local list only
                /// </summary>
                SV_TYPE_LOCAL_LIST_ONLY = 0x40000000,
                /// <summary>
                /// Lists available domains
                /// </summary>
                SV_TYPE_DOMAIN_ENUM = 0x80000000,
                SV_ALL = 0xffffffff, // TODO: only for testing
                // ReSharper restore InconsistentNaming
            }
            #endregion

            // Holds computer information
            [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
// ReSharper disable InconsistentNaming
            internal struct SERVER_INFO_101
// ReSharper restore InconsistentNaming
            {
                public int sv101_platform_id;
                public string sv101_name;
                public int sv101_version_major;
                public int sv101_version_minor;
                public int sv101_type;
                public string sv101_comment;
            }

            // enumerates network computers
            [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass")] // temporary
            [DllImport("Netapi32", CharSet = CharSet.Unicode), SuppressUnmanagedCodeSecurity]
            static private extern int NetServerEnum(
                string servername,		// must be null
                int level,				// 100 or 101
                out IntPtr bufptr,		// pointer to buffer receiving data
                int prefmaxlen,			// max length of returned data
                out int entriesread,	// num entries read
                out int totalentries,	// total servers + workstations
                uint servertype,		// server type filter
                [MarshalAs(UnmanagedType.LPWStr)]
			string domain,			// domain to enumerate
                IntPtr resume_handle);

            // Frees buffer created by NetServerEnum
            [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass")] // temporary
            [DllImport("netapi32.dll")]
            static private extern int NetApiBufferFree(
                IntPtr buf);

            // Constants
            private const int ERROR_ACCESS_DENIED = 5;
            private const int ERROR_MORE_DATA = 234;

            private NetworkComputers[] _computers;
            private string _lastError = "";

            /// <summary>
            /// Converts ServerType to its underlying value
            /// </summary>
            /// <param name="serverType">One of the ServerType values</param>
            /// <param name="domain">The domain to search for computers in</param>
            public CompEnum(ServerType serverType, string domain)
                : this(UInt32.Parse(Enum.Format(typeof(ServerType), serverType, "x"), NumberStyles.HexNumber), domain)
            {

            }
            /// <summary>
            /// Populates with broadcasting computers.
            /// </summary>
            /// <param name="serverType">Server type filter</param>
            /// <param name="domainName">The domain to search for computers in</param>
            public CompEnum(uint serverType, string domainName)
            {
                // Pointer to buffer that receives the data
                IntPtr pBuf = IntPtr.Zero;
                Type svType = typeof(SERVER_INFO_101);

                // structure containing info about the server
                SERVER_INFO_101 si;

                try
                {
                    var result = NetServerEnum(
                        null,
                        101,
                        out pBuf,
                        -1,
                        out int entriesread,    // number of entries actually read
                        out int totalentries,   // total visible servers and workstations
                        serverType,
                        domainName,
                        IntPtr.Zero);

                    // Successful?
                    if (result != 0)
                    {
                        switch (result)
                        {
                            case ERROR_MORE_DATA:
                                _lastError = "More data is available";
                                break;
                            case ERROR_ACCESS_DENIED:
                                _lastError = "Access was denied";
                                break;
                            default:
                                _lastError = "Unknown error code " + result;
                                break;
                        }
                        return;
                    }
                    else
                    {
                        _computers = new NetworkComputers[entriesread];

                        int tmp = (int)pBuf;
                        for (int i = 0; i < entriesread; i++)
                        {
                            // fill our struct
                            si = (SERVER_INFO_101)Marshal.PtrToStructure((IntPtr)tmp, svType);
                            _computers[i] = new NetworkComputers(si);

                            // next struct
                            tmp += Marshal.SizeOf(svType);
                        }
                    }
                }
                finally
                {
                    // free the buffer
                    NetApiBufferFree(pBuf);
                    pBuf = IntPtr.Zero;
                }
            }

            /// <summary>
            /// Total computers in the collection
            /// </summary>
            public int Length
            {
                get
                {
                    if (_computers != null)
                    {
                        return _computers.Length;
                    }
                    else
                    {
                        return 0;
                    }
                }
            }

            /// <summary>
            /// Last error message
            /// </summary>
            public string LastError => _lastError;

            /// <summary>
            /// Obtains the enumerator for ComputerEnumerator class
            /// </summary>
            /// <returns>IEnumerator</returns>
            public IEnumerator GetEnumerator()
            {
                return new ComputerEnumerator(_computers);
            }

            // cleanup
            [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1063:ImplementIDisposableCorrectly")] // temporary
            public void Dispose()
            {
                _computers = null;
            }

            public NetworkComputers this[int item] => _computers[item];

            // holds computer info.
            public struct NetworkComputers
            {
                SERVER_INFO_101 _computerinfo;
                internal NetworkComputers(SERVER_INFO_101 info)
                {
                    _computerinfo = info;
                }

                /// <summary>
                /// Name of computer
                /// </summary>
                public string Name => _computerinfo.sv101_name;

                /// <summary>
                /// Server comment
                /// </summary>
                public string Comment => _computerinfo.sv101_comment;

                /// <summary>
                /// Major version number of OS
                /// </summary>
                public int VersionMajor => _computerinfo.sv101_version_major;

                /// <summary>
                /// Minor version number of OS
                /// </summary>
                public int VersionMinor => _computerinfo.sv101_version_minor;
            }

            /// <summary>
            /// Enumerates the collection of computers
            /// </summary>
            public class ComputerEnumerator : IEnumerator
            {
                private NetworkComputers[] aryComputers;
                private int indexer;

                internal ComputerEnumerator(NetworkComputers[] networkComputers)
                {
                    aryComputers = networkComputers;
                    indexer = -1;
                }

                public void Reset()
                {
                    indexer = -1;
                }

                public object Current => aryComputers[indexer];

                public bool MoveNext()
                {
                    if (aryComputers == null)
                        return false;
                    if (indexer < aryComputers.Length)
                        indexer++;
                    return (!(indexer == aryComputers.Length));
                }
            }
        }

        public class BrowseableNetwork : IBrowseable
        {
            #region IBrowseable Members

            public void BrowseBranches(object parentNode, object nodeFilter, out IEnumerable<object> branches, out IEnumerable<Exception> exceptions)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Check.Ensures(Check.ValueAtReturn(out branches) != null);
                Check.Ensures(Check.ValueAtReturn(out exceptions) != null);

                if (parentNode is RootNetworkNode)
                {
//                    var cEnum = new CompEnum(CompEnum.ServerType.SV_TYPE_DOMAIN_ENUM, null);
                    var cEnum = new CompEnum(CompEnum.ServerType.SV_TYPE_DOMAIN_ENUM, null);
                    branches = CompEnumToEnum(cEnum, typeof(DomainNetworkNode));
                    exceptions = Enumerable.Empty<Exception>();
                    return;
                }
                if (parentNode is DomainNetworkNode domain)
                {
                    var cEnum = new CompEnum(CompEnum.ServerType.SV_ALL, domain.Name);
                    branches = CompEnumToEnum(cEnum, typeof(DomainNetworkNode));
                    exceptions = Enumerable.Empty<Exception>();
                    return;
                }
                throw new NotImplementedException();
            }

            public void BrowseLeaves(object parentNode, object nodeFilter, out IEnumerable<object> leaves, out IEnumerable<Exception> exceptions)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Check.Ensures(Check.ValueAtReturn(out leaves) != null);
                Check.Ensures(Check.ValueAtReturn(out exceptions) != null);

                leaves = Enumerable.Empty<Object>();
                exceptions = Enumerable.Empty<Exception>();
            }

            public void CreateNode(object parentNode, Type type, string name, out object node, out Exception exception)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Always.RequiresNotNull(type, nameof(type));
                Always.RequiresNotNull(name, nameof(name));
                Check.Ensures(Check.ValueAtReturn(out node) != null || Check.ValueAtReturn(out exception) != null);

                throw new NotImplementedException();
            }

            public IEnumerable<Type> GetAddableBranchTypes(object parentNode, bool onlyUserAddable)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Check.EnsuresResultNotNull<IEnumerable<Type>>();

                return Enumerable.Empty<Type>();
            }

            public IEnumerable<Type> GetAddableLeafTypes(object parentNode, bool onlyUserAddable)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Check.EnsuresResultNotNull<IEnumerable<Type>>();

                return Enumerable.Empty<Type>();
            }

            public void HasChildren(object parentNode, out bool parentHasBranches, out bool parentHasLeaves, out IEnumerable<Exception> exceptions)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Check.Ensures(Check.ValueAtReturn(out exceptions) != null);

                parentHasBranches = true;
                parentHasLeaves = false;
                exceptions = null;
            }

            public void InvalidateCache(object parentNode)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                //throw new NotImplementedException();
            }

            public bool IsBranch(object parentNode, object node)
            {
                Always.RequiresNotNull(node, nameof(node));
                return true;
                //throw new NotImplementedException();
            }

            public bool IsLeaf(object parentNode, object node)
            {
                Always.RequiresNotNull(node, nameof(node));
                return true;
                //throw new NotImplementedException();
            }

            public bool IsMyNode(object parentNode, object node)
            {
                Always.RequiresNotNull(node, nameof(node));
                return true;
            }

            public bool IsSelectable(object parentNode, object node)
            {
                Always.RequiresNotNull(node, nameof(node));
                return true;
                //throw new NotImplementedException();
            }

            #endregion

            private IEnumerable<Object> CompEnumToEnum(CompEnum compEnum, Type destType)
            {
                var list = new List<Object>();
                for (int i = 0; i < compEnum.Length; i++)
                {
                    switch (destType.Name)
                    {
                        case "DomainNetworkNode":
                            list.Add(new DomainNetworkNode(compEnum[i].Name));
                            break;
                    }
                }
                return list;
            }

            public class NamedNode
            {
                public NamedNode(string name)
                {
                    Name = name;
                }

                public string Name { get; private set; }
 
                public override string  ToString()
                {
 	                return Name ?? "";
                }
            }

            public class RootNetworkNode : NamedNode
            {
                public RootNetworkNode(string name)
                    : base(name)
                {
                }
            }

            public class DomainNetworkNode : NamedNode
            {
                public DomainNetworkNode(string name)
                    : base(name)
                {
                }
            }

            #region IBrowseable Members


            public bool MatchNode(object parentNode, object nodeA, object nodeB)
            {
                Always.RequiresNotNull(nodeA, nameof(nodeA));
                Always.RequiresNotNull(nodeB, nameof(nodeB));

                return Equals(nodeB, nodeA);
            }

            public void NodeAdded(object parentNode, object node)
            {
                Always.RequiresNotNull(parentNode, nameof(parentNode));
                Always.RequiresNotNull(node, nameof(node));
                // do nothing
            }

            #endregion
        }


        #endregion


        private readonly Timer _timer = new Timer { Interval = 1000 };
// ReSharper disable InconsistentNaming
#pragma warning disable IDE1006 // Naming Styles
        private void startButoton_Click(object sender, EventArgs e)
#pragma warning restore IDE1006 // Naming Styles
// ReSharper restore InconsistentNaming
        {
            _timer.Tick += timer_Tick;
            _timer.Start();
            if (listView1.SelectedItems.Count == 0)
            {
                listView1.Items.Add("I1");
                listView1.Items.Add("I2");
                listView1.Items.Add("I3");
            }
        }

// ReSharper disable InconsistentNaming
#pragma warning disable IDE1006 // Naming Styles
        void timer_Tick(object sender, EventArgs e)
#pragma warning restore IDE1006 // Naming Styles
// ReSharper restore InconsistentNaming
        {
            string focused = "Focused:";
            string selected = "Selected:";
            string lastSelected = "Last selected:";

            foreach (ListViewItem item in listView1.Items)
            {
                if (item.Focused) focused += " " + item;
                if (item.Selected) selected += " " + item;
            }
            if (listView1.SelectedItems.Count > 0) lastSelected += " " + listView1.SelectedItems[0];

            label1.Text = focused + Environment.NewLine + selected + Environment.NewLine + lastSelected;

        }

// ReSharper disable InconsistentNaming
#pragma warning disable IDE1006 // Naming Styles
        private void button4_Click(object sender, EventArgs e)
#pragma warning restore IDE1006 // Naming Styles
// ReSharper restore InconsistentNaming
        {
            listView1.Items[0].Selected = false;
        }

// ReSharper disable InconsistentNaming
#pragma warning disable IDE1006 // Naming Styles
        private void button5_Click(object sender, EventArgs e)
#pragma warning restore IDE1006 // Naming Styles
// ReSharper restore InconsistentNaming
        {
            listView1.Items[2].Focused = true;
        }

// ReSharper disable InconsistentNaming
#pragma warning disable IDE1006 // Naming Styles
        private void button6_Click(object sender, EventArgs e)
#pragma warning restore IDE1006 // Naming Styles
// ReSharper restore InconsistentNaming
        {
            if (listView1.FocusedItem != null) listView1.FocusedItem.Focused = false;
        }

// ReSharper disable InconsistentNaming
#pragma warning disable IDE1006 // Naming Styles
        private void listView1_SelectedIndexChanged(object sender, EventArgs e)
#pragma warning restore IDE1006 // Naming Styles
// ReSharper restore InconsistentNaming
        {
            if (listView1.Items[1].Selected) listView1.Items[1].Selected = false;
            if (listView1.Items[0].Selected) listView1.Items[0].Selected = false;
            if (listView1.Items[0].Focused) listView1.Items[0].Focused = false;
        }

// ReSharper disable InconsistentNaming
#pragma warning disable IDE1006 // Naming Styles
        private void button3_Click(object sender, EventArgs e)
#pragma warning restore IDE1006 // Naming Styles
// ReSharper restore InconsistentNaming
        {
            listView1.Items[0].Selected = true;
        }

        #region Test Browse Computers
        // For related or similar ideas or implementations, see also:
        // http://www.codeproject.com/Articles/16113/Retreiving-a-list-of-network-computer-names-using
        public sealed class NetworkBrowser
        {
            #region Dll Imports

            //declare the Netapi32 : NetServerEnum method import
            [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA2101:SpecifyMarshalingForPInvokeStringArguments", MessageId = "7")] // temporary
            [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA2101:SpecifyMarshalingForPInvokeStringArguments", MessageId = "0")] // temporary
            [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Interoperability", "CA1401:PInvokesShouldNotBeVisible")] // temporary
            [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass")] // temporary
            [DllImport("Netapi32", CharSet = CharSet.Auto,
            SetLastError = true),
            SuppressUnmanagedCodeSecurity]

            /// <summary>
            /// Netapi32.dll : The NetServerEnum function lists all servers
            /// of the specified type that are
            /// visible in a domain. For example, an 
            /// application can call NetServerEnum
            /// to list all domain controllers only
            /// or all SQL servers only.
            /// You can combine bit masks to list
            /// several types. For example, a value 
            /// of 0x00000003  combines the bit
            /// masks for SV_TYPE_WORKSTATION 
            /// (0x00000001) and SV_TYPE_SERVER (0x00000002)
            /// </summary>
            static public extern int NetServerEnum(
                string ServerNane, // must be null
                int dwLevel,
                ref IntPtr pBuf,
                int dwPrefMaxLen,
                out int dwEntriesRead,
                out int dwTotalEntries,
                int dwServerType,
                string domain, // null for login domain
                out int dwResumeHandle
                );

            //declare the Netapi32 : NetApiBufferFree method import
            [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Interoperability", "CA1401:PInvokesShouldNotBeVisible")] // temporary
            [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass")] // temporary
            [DllImport("Netapi32", SetLastError = true),
            SuppressUnmanagedCodeSecurity]

            /// <summary>
            /// Netapi32.dll : The NetApiBufferFree function frees 
            /// the memory that the NetApiBufferAllocate function allocates. 
            /// Call NetApiBufferFree to free
            /// the memory that other network 
            /// management functions return.
            /// </summary>
            static public extern int NetApiBufferFree(
                IntPtr pBuf);

            //create a _SERVER_INFO_100 STRUCTURE
            [StructLayout(LayoutKind.Sequential)]
            public struct _SERVER_INFO_100
            {
                internal int sv100_platform_id;
                [MarshalAs(UnmanagedType.LPWStr)]
                internal string sv100_name;
            }
            #endregion
            #region Public Constructor
            /// <SUMMARY>
            /// Constructor, simply creates a new NetworkBrowser object
            /// </SUMMARY>
            public NetworkBrowser()
            {

            }
            #endregion
            #region Public Methods
            /// <summary>
            /// Uses the DllImport : NetServerEnum
            /// with all its required parameters
            /// (see http://msdn.microsoft.com/library/default.asp?
            ///      url=/library/en-us/netmgmt/netmgmt/netserverenum.asp
            /// for full details or method signature) to
            /// retrieve a list of domain SV_TYPE_WORKSTATION
            /// and SV_TYPE_SERVER PC's
            /// </summary>
            /// <returns>Arraylist that represents
            /// all the SV_TYPE_WORKSTATION and SV_TYPE_SERVER
            /// PC's in the Domain</returns>
            public ArrayList getNetworkComputers()
            {
                //local fields
                ArrayList networkComputers = new ArrayList();
                const int MAX_PREFERRED_LENGTH = -1;
                int SV_TYPE_WORKSTATION = 1;
                int SV_TYPE_SERVER = 2;
                IntPtr buffer = IntPtr.Zero;
                IntPtr tmpBuffer = IntPtr.Zero;
                int entriesRead = 0;
                int totalEntries = 0;
                int resHandle = 0;
                int sizeofINFO = Marshal.SizeOf(typeof(_SERVER_INFO_100));


                try
                {
                    //call the DllImport : NetServerEnum 
                    //with all its required parameters
                    //see http://msdn.microsoft.com/library/
                    //default.asp?url=/library/en-us/netmgmt/netmgmt/netserverenum.asp
                    //for full details of method signature
                    int ret = NetServerEnum(null, 100, ref buffer,
                        MAX_PREFERRED_LENGTH,
                        out entriesRead,
                        out totalEntries, SV_TYPE_WORKSTATION |
                        SV_TYPE_SERVER, null, out 
                    resHandle);
                    //if the returned with a NERR_Success 
                    //(C++ term), =0 for C#
                    if (ret == 0)
                    {
                        //loop through all SV_TYPE_WORKSTATION 
                        //and SV_TYPE_SERVER PC's
                        for (int i = 0; i < totalEntries; i++)
                        {
                            //get pointer to, Pointer to the 
                            //buffer that received the data from
                            //the call to NetServerEnum. 
                            //Must ensure to use correct size of 
                            //STRUCTURE to ensure correct 
                            //location in memory is pointed to
                            tmpBuffer = new IntPtr((int)buffer +
                                       (i * sizeofINFO));
                            //Have now got a pointer to the list 
                            //of SV_TYPE_WORKSTATION and 
                            //SV_TYPE_SERVER PC's, which is unmanaged memory
                            //Needs to Marshal data from an 
                            //unmanaged block of memory to a 
                            //managed object, again using 
                            //STRUCTURE to ensure the correct data
                            //is marshalled 
                            _SERVER_INFO_100 svrInfo = (_SERVER_INFO_100)
                                Marshal.PtrToStructure(tmpBuffer,
                                        typeof(_SERVER_INFO_100));

                            //add the PC names to the ArrayList
                            networkComputers.Add(svrInfo.sv100_name);
                        }
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Problem with acessing " +
                        "network computers in NetworkBrowser " +
                        "\r\n\r\n\r\n" + ex.Message,
                        "Error", MessageBoxButtons.OK,
                        MessageBoxIcon.Error);
                    return null;
                }
                finally
                {
                    //The NetApiBufferFree function frees 
                    //the memory that the 
                    //NetApiBufferAllocate function allocates
                    NetApiBufferFree(buffer);
                }
                //return entries found
                return networkComputers;

            }
            #endregion
        }


        #region Test 2

        // For related or similar ideas or implementations, see also:
        //http://www.dreamincode.net/forums/topic/144747-list-network-computers/

        public class NetworkComputer
	{
	    public int computerID { get; set; }
	    public string computerName { get; set; }
	    public string computerIP { get; set; }
	    public List<SharedFolders> networkShares { get; set; }
	}
	 
	public class SharedFolders
	{
	    public string folderName { get; set; }
	    public string sharePath { get; set; }
	}
        //create a _SERVER_INFO_100 STRUCTURE
	[StructLayout(LayoutKind.Sequential)]
	public struct COMP_INFO
	{
	    internal int id;
	    [MarshalAs(UnmanagedType.LPWStr)]
	    internal string name;
	}
        //constants for LocalNetwork class
	public const int MAX_PREFERRED_LENGTH = -1;
	public const uint SV_TYPE_WORKSTATION = 1;
	public const uint SV_TYPE_SERVER = 2;
    public const uint SV_TYPE_ALL = 0xffffffff;
	public const uint SV_TYPE_NT = 0x00001000;
    public const uint SV_TYPE_WINDOWS = 0x00400000;
    public const uint SV_TYPE_DOMAIN_ENUM = 0x80000000;

        //declare the Netapi32 : NetServerEnum method import
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA2101:SpecifyMarshalingForPInvokeStringArguments", MessageId = "7")] // temporary
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Interoperability", "CA1401:PInvokesShouldNotBeVisible")] // temporary
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass")] // temporary
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Globalization", "CA2101:SpecifyMarshalingForPInvokeStringArguments", MessageId = "0")] // temporary
        [DllImport("Netapi32", CharSet = CharSet.Auto, SetLastError = true),
	SuppressUnmanagedCodeSecurity]
	static public extern int NetServerEnum(
	    string ServerNane, // must be null
	    int dwLevel,
	    ref IntPtr pBuf,
	    int dwPrefMaxLen,
	    out int dwEntriesRead,
	    out int dwTotalEntries,
	    uint dwServerType,
	    string domain, // null for login domain
	    out int dwResumeHandle
	    );

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Interoperability", "CA1401:PInvokesShouldNotBeVisible")] // temporary
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1060:MovePInvokesToNativeMethodsClass")] // temporary
        [DllImport("Netapi32", CharSet = CharSet.Auto, SetLastError = true, ExactSpelling = true)]

        static public extern int NetApiBufferFree(IntPtr Buffer);
        
        public List<NetworkComputer> getNetworkComputers()
        {
            //local fields
            List<NetworkComputer> computers = new List<NetworkComputer>();
            IntPtr buffer = IntPtr.Zero;
            IntPtr buffer2 = IntPtr.Zero;
            int numRead = 0;
            int count = 0;
            int handle = 0;
            int size = Marshal.SizeOf(typeof(COMP_INFO));


            try
            {
                int result = NetServerEnum(null, 100, ref buffer, MAX_PREFERRED_LENGTH, out numRead,
                    out count, SV_TYPE_DOMAIN_ENUM /*| SV_TYPE_WORKSTATION | SV_TYPE_SERVER | SV_TYPE_ALL | SV_TYPE_WINDOWS*/ , null, out  handle);

                //if the returned value is 0 (zero)
                if (result == 0)
                {
                    //now loop through everything returned
                    for (int i = 0; i < count; i++)
                    {
                        buffer2 = new IntPtr((int)buffer + (i * size));

                        COMP_INFO computer = (COMP_INFO)Marshal.PtrToStructure(buffer2, typeof(COMP_INFO));

                        try
                        {
                            //add the PC names to the list
                            computers.Add(
                                new NetworkComputer
                                {
                                    computerID = computer.id,
                                    computerName = computer.name,
                                    //computerIP = getIPAddress(computer.name)
                                    /*,
	                    networkShares = GetShares(computer.name)*/
                                });
                        }
                        catch
                        {
                            continue;
                        }


                    }
                }

                //this._isSuccessful = true;

                //return entries found
                return computers;

            }
            catch (Exception)
            {
                //this._message = ex.Message;
                //this._isSuccessful = false;
                return null;
            }
            finally
            {
                //free up memory
                NetApiBufferFree(buffer);
            }
        }

        #endregion

        #region Test 3
        public List<string> GetDomains()
        {
            var list = new List<string>();
            System.Management.SelectQuery query = new System.Management.SelectQuery("Win32_ComputerSystem");
            System.Management.ManagementObjectSearcher searcher = new System.Management.ManagementObjectSearcher(query);
            foreach (System.Management.ManagementObject mo in searcher.Get()) 
            {
                if((bool)mo["partofdomain"] != true)
                {
                    list.Add(String.Format("Workgroup {0} {1}", mo["domain"], mo["name"]));
                }
                else
                {
                    list.Add(String.Format("Domain {0} {1}", mo["domain"], mo["name"]));
                }
            }
            return list;
        }
        #endregion

        #region Test 4
        /*
        // For related or similar ideas or implementations, see also:
        //http://channel9.msdn.com/Forums/TechOff/Computer-names-on-network-c
        public List<String> ListNetworkComputers()
        {
                List<String> _ComputerNames = new List<String>();
                String _ComputerSchema="Computer";
                System.DirectoryServices.DirectoryEntry _WinNTDirectoryEntries = new System.DirectoryServices.DirectoryEntry("WinNT:");
                foreach (System.DirectoryServices.DirectoryEntry _AvailDomains in _WinNTDirectoryEntries.Children)
                {
                    foreach (System.DirectoryServices.DirectoryEntry _PCNameEntry in _AvailDomains.Children)
                    {
                        if (_PCNameEntry.SchemaClassName.ToLower().Contains(_ComputerSchema.ToLower()))
                        {
                            _ComputerNames.Add(_PCNameEntry.Name);
                        }
                    }
                }
            return _ComputerNames;
        }*/
        #endregion

        #region Test 5

        // For related or similar ideas or implementations, see also:
        //http://www.codeproject.com/script/Articles/ViewDownloads.aspx?aid=9390

        public List<string> GetComputers()
        {
            var list = new List<string>();
                         // Use Your work Group WinNT://&&&&(Work Group Name)
System.DirectoryServices.DirectoryEntry DomainEntry = new System.DirectoryServices.DirectoryEntry("WinNT://" + "CCDG");
//            System.DirectoryServices.DirectoryEntry DomainEntry = new System.DirectoryServices.DirectoryEntry("WinNT:");
//DomainEntry.Children.SchemaFilter.Add("computer");
              
  
                  // To Get all the System names And Display with the Ip Address
            foreach(System.DirectoryServices.DirectoryEntry machine in DomainEntry.Children)
            {
                list.Add(machine.Name);
            }
            return list;
        }
        #endregion

        #region Test 6

        // For related or similar ideas or implementations, see also:
        //http://stackoverflow.com/questions/5964390/how-to-find-out-if-a-computer-is-connected-to-novell-edirectory-or-microsoft-act

        public List<string> GetTest6()
        {
            var list = new List<string>();
            System.Management.ManagementObjectSearcher domainInfos = new System.Management.ManagementObjectSearcher("select * from WIN32_NTDomain");

            foreach (System.Management.ManagementObject domainInfo in domainInfos.Get())
            {
                list.Add(String.Format("Name : {0}", domainInfo.GetPropertyValue("Name")));
                list.Add(String.Format("Computer/domain : {0}", domainInfo.GetPropertyValue("Caption")));
                list.Add(String.Format("Domain name : {0}", domainInfo.GetPropertyValue("DomainNAme")));
                list.Add(String.Format("Status : {0}", domainInfo.GetPropertyValue("Status")));
            }
            return list;
        }

        #endregion

#pragma warning disable IDE1006 // Naming Styles
        private void getComputersButton_Click(object sender, EventArgs e)
#pragma warning restore IDE1006 // Naming Styles
        {
            int test = 2;
            switch (test)
            {
                case 1:
                    NetworkBrowser nb1 = new NetworkBrowser();
                    foreach (string pc in nb1.getNetworkComputers())
                    {
                       listView1.Items.Add(pc); // if dont FW returns wxppcz01, ale pouze z vlastni workgroup
                    }
                    break;

                case 2:
            
                    var nb2 = getNetworkComputers();
                    if (nb2 != null)
                    foreach (NetworkComputer pc in nb2)
                    {
                        listView1.Items.Add(pc.computerName); // if dont FW:  DOMAIN: CCDG, if WINDOWS: wxppcz01
                    }
                    break;

                case 3:
                    foreach (string s in GetDomains()) listView1.Items.Add(s); // returns CCDG/CCD01 - only this computer
                    break;

                case 5:
                    foreach (string s in GetComputers()) listView1.Items.Add(s); // if dont FW, returns wxppcz01 and schema - jen z vlastni workgroup
                    break;

                case 6:
                    foreach (string s in GetTest6()) listView1.Items.Add(s); // returns CCD01, CCD01 2x - only local PC
                    break;

                case 7:
                    listView1.Items.Add(Environment.MachineName);
                    break;
        }
        }

        #endregion

        #region Test PropertyGrid

        //[TypeConverter(typeof(ExpandableObjectConverter))]
        [TypeConverter(typeof(InsideClassConverter))] // ***** this must be for bold only other value than default value defined on parent class
        public class InsideClass
        {
            public InsideClass(string name)
            {
                Name = name;
            }

            [DefaultValue("Unknown")]
            [RefreshProperties(RefreshProperties.All)]  // ***** this must be for bold only other value than default value defined on parent class
            public string Name { get; set; }

            /// <inheritdoc cref="object.ToString()"/>
            public override string ToString() // ***** this must be for bold only other value than default value defined on parent class
            {
                return Name ?? "";
            }

            /// <inheritdoc cref="Object.Equals(object)"/>
            public override bool Equals(object obj)
            // ***** this must be for bold only other value than default value defined on parent class
            {
#pragma warning disable IDE0019 // Use pattern matching
                var insideClass = obj as InsideClass;
#pragma warning restore IDE0019 // Use pattern matching
                if (insideClass != null) return String.Equals(Name, insideClass.Name);
                return base.Equals(obj);
            }

            /// <inheritdoc cref="Object.GetHashCode"/>
            public override int GetHashCode()
            {
                return base.GetHashCode() ^ Name.GetHashCode();
            }
        }

        private class OutsideClass
        {
            public int Id { get; set; }

            [Category("Test")]
            [DefaultValue(typeof(InsideClass), "Unknown")] // ***** this must be for bold only other value than default value defined on parent class
            //            [DefaultValue("Check while typing:Unknown")]
            //[DesignerSerializationVisibility(DesignerSerializationVisibility.Content)] 
            [PropertyGridInitialExpanded(true)] // thid is for auto expand after call propertyGrid.ExpandItemWithInitialExpandedAttribute()
            public InsideClass Name
            {
                get => _name;
                set => _name = value;
            }

            private InsideClass _name = new InsideClass("Unknown");

            [Category("Test")]
            [DefaultValue(typeof(InsideClass), "Unknown")] // ***** this must be for bold only other value than default value defined on parent class
            public InsideClass SecondName
                        {
                            get => _secondName;
                set => _secondName = value;
            }

            private InsideClass _secondName = new InsideClass("Unknown2");

            /*
            // after uncomment this default value not bold - not right
            private bool ShouldSerializeName()
            {
                if (String.Equals(Name.Name, "Unknown")) return false;
                return true;
            }*/
        }

        public class InsideClassConverter : ExpandableObjectConverter
        {
            /*
            public override bool CanConvertTo([CanBeNull] ITypeDescriptorContext context, Type destinationType)
            {
                if (destinationType == typeof(InsideClass))
                    return true;

                return base.CanConvertTo(context, destinationType);
            }*/

            private bool expandDone = false;
            // this method allow display other value instead of "ToString"
            public override object ConvertTo(
                [CanBeNull] ITypeDescriptorContext context,
                [CanBeNull] CultureInfo culture,
                [CanBeNull] object value,
                Type destinationType)
            {
                if (destinationType == typeof(String) && value is InsideClass)
                {
                    if (context != null)
                    {
                        if (!expandDone)
                        {
                            PropertyInfo piExpanded = context.GetType().GetProperty("Expanded");
                            //Object expanded = piExpanded.GetValue(context, null);
                            piExpanded.SetValue(context, true, null);
                            expandDone = true;
                        }
                    }

                    InsideClass so = (InsideClass)value;

                    return "Check while typing:" + so.Name;
                }
                return base.ConvertTo(context, culture, value, destinationType);
            }
            
            
            // this method allow editing "ToString()" value in property grid
            public override bool CanConvertFrom(ITypeDescriptorContext context,
                              Type sourceType)
            {
                if (sourceType == typeof(string))
                    return true;

                return base.CanConvertFrom(context, sourceType);
            }

            public override object ConvertFrom(ITypeDescriptorContext context, CultureInfo culture, object value)
            // ***** this must be for bold only other value than default value defined on parent class
            {
                if (value is string)
                {
                    try
                    {
                        string s = (string)value;
                        int colon = s.IndexOf(':');

                        if (colon != -1)
                        {
                            string suggCorr = s.Substring(colon + 1);

                            InsideClass so = new InsideClass(suggCorr);

                            return so;
                        }
                        return new InsideClass(s);
                    }
                    catch
                    {
                        throw new ArgumentException(
                            "Can not convert '" + (string)value +
                                               "' to type InsideClass");
                    }
                }
                return base.ConvertFrom(context, culture, value);
            }

            
        }





// ReSharper disable InconsistentNaming
#pragma warning disable IDE1006 // Naming Styles
        private void testGridbutton_Click(object sender, EventArgs e)
#pragma warning restore IDE1006 // Naming Styles
// ReSharper restore InconsistentNaming
        {
            propertyGrid1.SelectedObject = new OutsideClass();
            //propertyGrid1.ExpandItemWithInitialExpandedAttribute(); // auto expand property with attribute [PropertyGridInitialExpanded(true)]
            //ExpandPropertyGridItem(propertyGrid1, "Name"); // auto expand property Name
        }

        /// <summary>
        /// Expands first property which has label.
        /// </summary>
        /// <param name="propertyGrid"> </param>
        /// <param name="label"> </param>
        /// <param name="gridItem">Searching starts at grid item. If gridItem is null, start from root.</param>
        static private void ExpandPropertyGridItem(PropertyGrid propertyGrid, string label, GridItem gridItem = null)
        {
            if (propertyGrid == null) return;
            if (gridItem == null)
            { // find from root
                if (propertyGrid.SelectedGridItem != null)
                {
                    gridItem = propertyGrid.SelectedGridItem.Parent;
                    while (gridItem.Parent != null) gridItem = gridItem.Parent;
                }
            }

            if (gridItem != null)
            {
                if (gridItem.GridItemType == GridItemType.Property && gridItem.Expandable)
                {
                    if (gridItem.Label == label)
                    {
                        GridItem gi = gridItem;
                        while (gi != null)
                        {
                            gi.Expanded = true;
                            gi = gi.Parent;
                        }
                        return;
                    }
                }

                foreach (GridItem childItem in gridItem.GridItems)
                {
                    ExpandPropertyGridItem(propertyGrid, label, childItem);
                }
            }


        }






        #endregion

#pragma warning disable IDE1006 // Naming Styles
        private void listView1_DoubleClick(object sender, EventArgs e)
#pragma warning restore IDE1006 // Naming Styles
        {

        }

#pragma warning disable IDE1006 // Naming Styles
        private void button7_Click(object sender, EventArgs e)
#pragma warning restore IDE1006 // Naming Styles
        {
            ShowBrowseNodes(true);
        }

#pragma warning disable IDE1006 // Naming Styles
        private void showDAItemDialogButton_Click(object sender, EventArgs e)
#pragma warning restore IDE1006 // Naming Styles
        {
            //daItemDialog1.ServerDescriptor = "OPCLabs.KitServer.2";
            //daItemDialog1.ShowDialog();
        }

        
    }


    static public class PropertyGridExtension
    {
        /*
        public static void ExpandGroupName(this PropertyGrid propertyGrid, string groupName)
        {
            foreach (GridItem gridItem in propertyGrid.SelectedGridItem.GridItems) // this is not working, must find root
            {
                if (gridItem != null)
                {
                    if (gridItem.GridItemType == GridItemType.Category && gridItem.Label == groupName)
                    {
                        gridItem.Expanded = true;
                    }
                }
            }
        }*/

        static public void ExpandItemWithInitialExpandedAttribute(this PropertyGrid propertyGrid)
        {
            if (propertyGrid.SelectedGridItem != null)
            {
                GridItem gridItem = propertyGrid.SelectedGridItem;
                while (gridItem.Parent != null) gridItem = gridItem.Parent;

                ExpandItemWithInitialExpandedAttribute(propertyGrid, gridItem);
            }
        }

        static private void ExpandItemWithInitialExpandedAttribute(PropertyGrid propertyGrid, GridItem gridItem)
        {
            if (gridItem != null)
            {
                if (gridItem.GridItemType == GridItemType.Property && gridItem.Expandable)
                {
                    foreach (Attribute attribute in gridItem.PropertyDescriptor.Attributes)
                    {
                        if (attribute is PropertyGridInitialExpandedAttribute attr)
                        {
                            if (attr.InitialExpanded)
                            {
                                GridItem gi = gridItem;
                                while (gi != null)
                                {
                                    gi.Expanded = true;
                                    gi = gi.Parent;
                                }

                            }
                            break;
                        }
                    }
                }

                foreach (GridItem childItem in gridItem.GridItems)
                {
                    ExpandItemWithInitialExpandedAttribute(propertyGrid, childItem);
                }
            }
        }

    }

    public class PropertyGridInitialExpandedAttribute : Attribute
    {
        public bool InitialExpanded { get; set; }

        public PropertyGridInitialExpandedAttribute(bool initialExpanded)
        {
            InitialExpanded = initialExpanded;
        }
    }

}
