﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Specialized;
using OpcLabs.BaseLib.OperationModel;
using OpcLabs.BaseLib.Utilities;
using OpcLabs.BaseLibFactory.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.BaseLibTest.OperationModel
{
    public partial class OperationResultTest
    {
        #region Test Methods
        // ReSharper disable InconsistentNaming

        [TestMethod]
        public void BinarySerializationTest_RoundTrip_Instance2()
        {
            Tests.BinarySerializationTest(
                OperationResultInstances.Instance2, 
                (arg1, arg2) => Assert.IsTrue(ObjectUtilities.ValueEquals(arg1, arg2)));
        }

        [TestMethod]
        public void DataContractSerializationTest_RoundTrip_Instance2()
        {
            Tests.DataContractSerializationTest(
                OperationResultInstances.Instance2,
                (arg1, arg2) => Assert.IsTrue(ObjectUtilities.ValueEquals(arg1, arg2)));
        }

        [TestMethod]
        public void Formatting_F_Failure_ErrorId_Test()
        {
            OperationResult target = new OperationResult { ErrorMessage = "Some error\nSecond line", ErrorId = "-2" };
            Assert.AreEqual("*** Failure -2: Some error [...] (0 diagnostics)", target.ToString("F"));
        }

        [TestMethod]
        public void Formatting_F_Failure_NoErrorId_Test()
        {
            Assert.AreEqual("*** Failure System.Exception: Some error [...] (0 diagnostics)", new OperationResult { ErrorMessage = "Some error\nSecond line" }.ToString("F"));
        }

        [TestMethod]
        public void Formatting_F_Success_Diagnostics_Test()
        {
            Assert.AreEqual(
                "Success (4 diagnostics)",
                new OperationResult
                {
                    Diagnostics = new ExceptionCollection(Enumerable.Repeat<Exception>(new ApplicationException(), 4).ToList())
                }.ToString("F"));
        }

        [TestMethod]
        public void Formatting_F_Success_NoDiagnostics_Test()
        {
            Assert.AreEqual("Success (0 diagnostics)", new OperationResult().ToString("F"));
        }

        [TestMethod]
        public void Formatting_g_Failure_ErrorId_Test()
        {
            Assert.AreEqual("*** Failure -2: Some error [...]", new OperationResult { ErrorMessage = "Some error\nSecond line", ErrorId = "-2" }.ToString("g"));
        }

        [TestMethod]
        public void Formatting_g_Failure_NoErrorId_Test()
        {
            Assert.AreEqual("*** Failure System.Exception: Some error [...]", new OperationResult { ErrorMessage = "Some error\nSecond line" }.ToString("g"));
        }

        [TestMethod]
        public void Formatting_g_Success_Diagnostics_Test()
        {
            Assert.AreEqual(
                "Success",
                new OperationResult
                {
                    Diagnostics = new ExceptionCollection(Enumerable.Repeat<Exception>(new ApplicationException(), 4).ToList())
                }.ToString("g"));
        }

        [TestMethod]
        public void Formatting_g_Success_NoDiagnostics_Test()
        {
            Assert.AreEqual("Success", new OperationResult().ToString("g"));
        }

        [TestMethod]
        public void ToString_Failure_ErrorId_Test()
        {
            Assert.AreEqual("*** Failure -2: Some error [...]", new OperationResult { ErrorMessage = "Some error\nSecond line", ErrorId = "-2" }.ToString());
        }

        [TestMethod]
        public void ToString_Failure_NoErrorId_Test()
        {
            Assert.AreEqual("*** Failure System.Exception: Some error [...]", new OperationResult { ErrorMessage = "Some error\nSecond line" }.ToString());
        }

        [TestMethod]
        public void ToString_Success_Diagnostics_Test()
        {
            Assert.AreEqual(
                "Success (4 diagnostics)",
                new OperationResult
                {
                    Diagnostics = new ExceptionCollection(Enumerable.Repeat<Exception>(new ApplicationException(), 4).ToList())
                }.ToString());
        }

        [TestMethod]
        public void ToString_Success_NoDiagnostics_Test()
        {
            Assert.AreEqual("Success", new OperationResult().ToString());
        }

        // ReSharper restore InconsistentNaming
        #endregion
    }
}
