﻿using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;

namespace OpcLabs.EasyOpcLiveTest.DataAccess
{
    // ReSharper disable once InconsistentNaming
    public partial class IEasyDAClientTest_Intrinsic : IEasyDAClientTestBase
    {
        #region Public Constructors

        public IEasyDAClientTest_Intrinsic([NotNull] Func<IEasyDAClient> createClientFunc)
            : base(createClientFunc)
        {
        }

        #endregion



        #region Test Methods

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseServersTest_EmptyString_NoTechnology()
        {
            BrowseServersTest_NoTechnology("");
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseServersTest_LocalHost_NoTechnology()
        {
            BrowseServersTest_NoTechnology("localhost");
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ConstructionTest()
        {
            var client = CreateClient();
            ClientAsserts(client);
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ConstructionTest_Repeated_10000()
        {
            const int count = 10000;
            for (int i = 0; i < count; i++)
                ConstructionTest();
        }

        #endregion



        #region Protected Methods

        protected virtual void ClientAsserts([CanBeNull] IEasyDAClient client)
        {
            Assert.IsNotNull(client);
        }

        protected void BrowseServersTest_NoTechnology([NotNull] string location)
        {
            IEasyDAClient client = CreateClient();
            ServerElementCollection serverElements = client.BrowseServers(location, OpcTechnologies.None);
            Assert.IsNotNull(serverElements);
            Assert.AreEqual(0, serverElements.Count);
        }

        #endregion
    }
}
