﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Generic.Utilities;
using OpcLabs.BaseLib.OperationModel.Generic;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.AddressSpace.Standard;
using OpcLabs.EasyOpc.UA.DataTypeModel.Implementation;
using OpcLabs.EasyOpc.UA.Extensions;
using OpcLabs.EasyOpc.UA.InformationModel;
using OpcLabs.EasyOpc.UA.Navigation;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpc.UA.Operations;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.UA.DataTypeModel
{
    [TestClass]
    // ReSharper disable once InconsistentNaming
    public class DataTypeDictionaryTest_Experiments
    {
        #region Static Private Methods

        [NotNull]
        static private EasyUAClient CreateClientWithoutComplexDataPlugin()
        {
            var client = new EasyUAClient();
            // ReSharper disable once PossibleNullReferenceException
            client.InstanceParameters.PluginSetups[EasyOpcUAUriStrings.ComplexDataPlugin].Enabled = false;
            return client;
        }

        #endregion



        #region Test Methods

        [TestMethod]
        public void DumpDataTypesTest_OpcFoundation_ReferenceServer() =>
            DumpDataTypesTest(UALiveTestEnvironment.OpcFoundation_UAReferenceServer.EndpointDescriptors.OpcTcp);

        [TestMethod]
        public void DumpDataTypesTest_OpcFoundation_SampleServer() =>
            DumpDataTypesTest(UALiveTestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp);

        [TestMethod]
        public void DumpDataTypesTest_ProsysOpc_SimulationServer() =>
            DumpDataTypesTest(UALiveTestEnvironment.Prosys_UASimulationServer.EndpointDescriptors.OpcTcp);

        [TestMethod]
        public void DumpDataTypesTest_Softing_OpcUaNetDemoServer() =>
            DumpDataTypesTest(UALiveTestEnvironment.Softing_OpcUaNetDemoServer.EndpointDescriptors.OpcTcp);

        [TestMethod]
        public void DumpDataTypesTest_UnifiedAutomation_UaCppServer() =>
            DumpDataTypesTest(UALiveTestEnvironment.UnifiedAutomation_UaCppServer.EndpointDescriptors.OpcTcp);

        [TestMethod]
        public void ProofOfConceptTest1()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UALiveTestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            AssertUtilities.WriteLine($"endpointDescriptor: {endpointDescriptor}");
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);
            
            UANodeDescriptor nodeDescriptor = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_Scalar_StructureValue;

            var client = CreateClientWithoutComplexDataPlugin();

            TestSection.Act();

            object value = client.ReadValue(endpointDescriptor, nodeDescriptor);
            AssertUtilities.WriteLine($"value: {value}");
            Assert.IsNotNull(value);
            var extensionObject = value as UAExtensionObject;
            Assert.IsNotNull(extensionObject);
            byte[] binaryBody = extensionObject.BinaryBody;
            Assert.IsNotNull(binaryBody);
            AssertUtilities.WriteLine($"binaryBody: {BitConverter.ToString(binaryBody)}");

            // Note that we have internal ReadDataType for this
            UANodeId dataTypeId = (UANodeId)client.ReadValue(endpointDescriptor, nodeDescriptor, UAAttributeId.DataType);
            AssertUtilities.WriteLine($"dataTypeID: {dataTypeId}");
            Assert.IsNotNull(dataTypeId);

            UANodeElementCollectionResult nodeElementCollectionResult = client.RecursiveBrowse(
                new UABrowseArguments(endpointDescriptor, dataTypeId, new UABrowseParameters(
                    UANodeClass.DataType, new[] {UAReferenceTypeIds.HasSubtype}, includeSubtypes: true)));
            Assert.IsNotNull(nodeElementCollectionResult);
            UANodeElementCollection nodeElementCollection = nodeElementCollectionResult.Value;
            Assert.IsNotNull(nodeElementCollection);

            IEnumerable<UANodeId> nodeIds = nodeElementCollection
                // ReSharper disable once PossibleNullReferenceException
                .Select(element => element.NodeId)
                .Concat(new [] {dataTypeId});

            UAReadArguments[] readArgumentsArray = nodeIds
                // ReSharper disable once AssignNullToNotNullAttribute
                .Select(id => new UAReadArguments(endpointDescriptor, id, UAAttributeId.IsAbstract)).ToArray();
            UAAttributeDataResult[] readResultArray = client.ReadMultiple(readArgumentsArray);




            UANodeElementCollection encodingNodeCollection = client.Browse(endpointDescriptor, dataTypeId,
                new UABrowseParameters(
                    UANodeClass.All,
                    EnumerableUtilities.One(UAReferenceTypeIds.HasEncoding), 
                    includeSubtypes: true));
            Assert.IsNotNull(encodingNodeCollection);

            // unfortunately this gives no references...
        }

        // gives best results so far
        [TestMethod]
        public void ProofOfConceptTest2_UaCppServer()
        {
            // UaCppServer (Unified Automation)
            UAEndpointDescriptor endpointDescriptor = UALiveTestEnvironment.UnifiedAutomation_UaCppServer.EndpointDescriptors.OpcTcp;
            UANodeDescriptor nodeDescriptor = "nsu=urn:UnifiedAutomation:CppDemoServer:BuildingAutomation;s=ControllerConfigurations";

            ProofOfConceptTest2(endpointDescriptor, nodeDescriptor);
        }

        // has the node exists but returns a null value, and data type is an abstract Structure
        //[TestMethod]
        //public void ProofOfConceptTest2_QuickStarts_ReferenceServer()
        //{
        //    // UaCppServer (Unified Automation)
        //    UAEndpointDescriptor endpointDescriptor = OpcFoundation_ReferenceServer.EndpointDescriptor_OpcTcp_TestingInsecure;
        //    UANodeDescriptor nodeDescriptor = "nsu=http://opcfoundation.org/Quickstarts/ReferenceApplications;s=DataAccess_DataItem_ExtensionObject";

        //    ProofOfConceptTest2(endpointDescriptor, nodeDescriptor);
        //}

        #endregion



        #region Private Methods

        private void DumpDataTypesTest([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            TestSection.Arrange();
            AssertUtilities.WriteLine($"endpointDescriptor: {endpointDescriptor}");
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            AssertUtilities.WriteLine();
            var client = new EasyUAClient();

            TestSection.Act();

            ValueResult<UANodeDescriptor[]> dataTypeArrayResult = client.RetrieveAllDataTypes(endpointDescriptor);
            Assert.IsNotNull(dataTypeArrayResult);
            AssertUtilities.WriteLine(dataTypeArrayResult);
            Assert.IsTrue(dataTypeArrayResult.Succeeded);
            UANodeDescriptor[] dataTypes = dataTypeArrayResult.Value;
            Assert.IsNotNull(dataTypes);

            int count = dataTypes.Length;
            AssertUtilities.WriteLine(count);
            foreach (UANodeDescriptor dataType in dataTypes)
            {
                AssertUtilities.WriteLine(dataType);

                //var x = client.Browse(endpointDescriptor, dataType,
                //    new UABrowseParameters(UANodeClass.All, UABrowseDirections.Both, new [] {UAReferenceTypeIds.References}));

            }

            var dataTypeModelProvider = new UAServerDataTypeModelProvider(new UAOperationFactory(() => client));
            var dataTypeDictionaryProvider = new UAServerDataTypeDictionaryProvider(new UAOperationFactory(() => client));

            ValueResult<UANodeElementCollection>[] encodingNodeElementCollections = 
                dataTypeModelProvider.GetEncodingNodeElementCollectionsFromDataTypeIds(
                    // ReSharper disable once AssignNullToNotNullAttribute
                    dataTypes.Select(element => new UAModelNodeDescriptor(endpointDescriptor, element)).ToList());
            Assert.IsNotNull(encodingNodeElementCollections);

            AssertUtilities.WriteLine();
            var dataTypeDictionaryIdSet = new HashSet<UAModelNodeDescriptor>();
            for (int i = 0; i < count; i++)
            {
                ValueResult<UANodeElementCollection> result = encodingNodeElementCollections[i];
                Assert.IsNotNull(result);
                Assert.IsTrue(result.Succeeded);

                UANodeElementCollection nodeElementCollection = result.Value;
                Assert.IsNotNull(nodeElementCollection);

                if (result.Succeeded && (nodeElementCollection.Count != 0))
                {
                    AssertUtilities.WriteLine();
                    UANodeDescriptor dataTypeId = dataTypes[i];
                    AssertUtilities.WriteLine($"dataTypeId: {dataTypeId} ({nodeElementCollection.Count} encodingIds)");
                    Assert.IsNotNull(dataTypeId);
                    AssertUtilities.WriteLine($"dataTypeId.NodeID: {dataTypeId.NodeId}");
                    foreach (UANodeElement nodeElement in nodeElementCollection)
                    {
                        Assert.IsNotNull(nodeElement);

                        AssertUtilities.WriteLine();
                        AssertUtilities.WriteLine($"  encodingId: {nodeElement}");
                        ValueResult<UAModelNodeDescriptor>[] dataTypeDescriptionIdResultArray = dataTypeDictionaryProvider.GetDataTypeDescriptionIdsFromDataTypeEncodingIds(new[]
                            {new UAModelNodeDescriptor(endpointDescriptor, nodeElement)});
                        Assert.IsNotNull(dataTypeDescriptionIdResultArray);
                        ValueResult<UAModelNodeDescriptor> dataTypeDescriptionIdResult = dataTypeDescriptionIdResultArray[0];
                        Assert.IsNotNull(dataTypeDescriptionIdResult);
                        UAModelNodeDescriptor dataTypeDescriptionId = dataTypeDescriptionIdResult.Value;
                        AssertUtilities.WriteLine($"    dataTypeDescriptionId: {dataTypeDescriptionId}");
                        Assert.IsNotNull(dataTypeDescriptionId);

                        ValueResult<UAModelNodeDescriptor>[] dataTypeDictionaryIdResultArray = dataTypeDictionaryProvider.GetDataTypeDictionaryIdsFromDataTypeDescriptionIds(
                            // ReSharper disable once PossibleNullReferenceException
                                dataTypeDescriptionIdResultArray.Select(valueResult => valueResult.Value).ToArray());
                        Assert.IsNotNull(dataTypeDictionaryIdResultArray);
                        ValueResult<UAModelNodeDescriptor> dataTypeDictionaryIdResult = dataTypeDictionaryIdResultArray[0];
                        Assert.IsNotNull(dataTypeDictionaryIdResult);
                        UAModelNodeDescriptor dataTypeDictionaryId = dataTypeDictionaryIdResult.Value;
                        AssertUtilities.WriteLine($"    dataTypeDictionaryId: {dataTypeDictionaryId}");
                        Assert.IsNotNull(dataTypeDictionaryId);

                        dataTypeDictionaryIdSet.Add(dataTypeDictionaryId);
                    }



                }
            }

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine($"Number of distinct data type dictionary IDs: {dataTypeDictionaryIdSet.Count}");

            foreach (UAModelNodeDescriptor dataTypeDictionaryId in dataTypeDictionaryIdSet)
            {
                Debug.Assert(dataTypeDictionaryId != null);
                AssertUtilities.WriteLine();
                AssertUtilities.WriteLine($"dataTypeDictionaryId: {dataTypeDictionaryId}");

                ValueResult<byte[]>[] dataTypeDictionaryResultArray = dataTypeDictionaryProvider.GetDataTypeDictionariesFromDataTypeDictionaryIds(
                    new[] { dataTypeDictionaryId });
                Assert.IsNotNull(dataTypeDictionaryResultArray);

                ValueResult<byte[]> dataTypeDictionaryResult = dataTypeDictionaryResultArray[0];
                Assert.IsNotNull(dataTypeDictionaryResult);
                Assert.IsTrue(dataTypeDictionaryResult.Succeeded);
                var dictionaryArrayOfBytes = dataTypeDictionaryResult.Value;
                if (dictionaryArrayOfBytes == null)
                    continue;

                string dictionaryString = Encoding.UTF8.GetString(dictionaryArrayOfBytes);
                AssertUtilities.WriteLine("dictionaryString:");
                AssertUtilities.WriteLine(dictionaryString);
            }
        }

        private void ProofOfConceptTest2(
            [NotNull] UAEndpointDescriptor endpointDescriptor, 
            [NotNull] UANodeDescriptor nodeDescriptor)
        {
            TestSection.Arrange();

            AssertUtilities.WriteLine($"endpointDescriptor: {endpointDescriptor}");
            AssertUtilities.WriteLine($"nodeDescriptor: {nodeDescriptor}");

            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            AssertUtilities.WriteLine();
            var client = CreateClientWithoutComplexDataPlugin();

            TestSection.Act();

            AssertUtilities.WriteLine();
            object value = client.ReadValue(endpointDescriptor, nodeDescriptor);
            AssertUtilities.WriteLine($"value: {value}");
            if (value != null)
            {
                AssertUtilities.WriteLine($"value.GetType(): {value.GetType()}");
                var extensionObject = value as UAExtensionObject;
                Assert.IsNotNull(extensionObject);
                AssertUtilities.WriteLine($"extensionObject.DataTypeEncodingId: {extensionObject.DataTypeEncodingId}");
                AssertUtilities.WriteLine($"extensionObject.ExtensionObjectType: {extensionObject.ExtensionObjectType}");
                object body = extensionObject.Body;
                Assert.IsNotNull(body);
                AssertUtilities.WriteLine($"body.GetType(): {body.GetType()}");
                AssertUtilities.WriteLine($"body: {body}");
                var bodyArrayOfBytes = body as byte[];
                Assert.IsNotNull(bodyArrayOfBytes);
                AssertUtilities.WriteLine($"bodyArrayOfBytes: {BitConverter.ToString(bodyArrayOfBytes)}");
            }

            AssertUtilities.WriteLine();
            UANodeId dataTypeId = client.ReadDataType(endpointDescriptor, nodeDescriptor);
            Assert.IsNotNull(dataTypeId);
            AssertUtilities.WriteLine($"dataTypeId: {dataTypeId}");

            AssertUtilities.WriteLine();
            // ReSharper disable once PossibleNullReferenceException
            bool isAbstract = (bool)client.ReadValue(endpointDescriptor, dataTypeId, UAAttributeId.IsAbstract);
            AssertUtilities.WriteLine($": {isAbstract}");
            Assert.IsFalse(isAbstract);

            AssertUtilities.WriteLine();
            // ReSharper disable once PossibleNullReferenceException
            var valueRank = (int) client.ReadValue(endpointDescriptor, nodeDescriptor, UAAttributeId.ValueRank);
            Assert.IsNotNull(valueRank);
            AssertUtilities.WriteLine($"valueRank: {valueRank}");

            AssertUtilities.WriteLine();
            UANodeElementCollection encodingNodeCollection = client.Browse(endpointDescriptor, dataTypeId,
                new UABrowseParameters(
                    UANodeClass.All, // may restrict more
                    EnumerableUtilities.One(UAReferenceTypeIds.HasEncoding),
                    includeSubtypes: true));
            Assert.IsNotNull(encodingNodeCollection);
            AssertUtilities.WriteLine($"encodingNodeCollection.Count: {encodingNodeCollection.Count}");
            Assert.AreEqual(1, encodingNodeCollection.Count); // may have more nodes
            UANodeElement encodingElement = encodingNodeCollection[0];
            Assert.IsNotNull(encodingElement);
            AssertUtilities.WriteLine($"encodingElement: {encodingElement}");
            var browseName = encodingElement.BrowseName;
            AssertUtilities.WriteLine($"browseName: {browseName}");
            AssertUtilities.WriteLine($"browseName.NamespaceUriString: {browseName.NamespaceUriString}");
            AssertUtilities.WriteLine($"browseName.Name: {browseName.Name}");
            Assert.AreEqual("http://opcfoundation.org/UA/", browseName.NamespaceUriString);
            Assert.AreEqual("Default Binary", browseName.Name);
            UANodeId dataTypeEncodingId = encodingElement.NodeId;
            Assert.IsNotNull(dataTypeEncodingId);
            AssertUtilities.WriteLine($"dataTypeEncodingId: {dataTypeEncodingId}"); // DataTypeEncoding object
            // this is the same as extensionObject.DataTypeEncodingId, i.e. can be found directly from the value!

            // gives BadAttributeIdInvalid
            //AssertUtilities.WriteLine();
            //object encoding = client.ReadValue(endpointDescriptor, encodingId);
            //Assert.IsNotNull(encoding);
            //AssertUtilities.WriteLine($"encoding.GetType(): {encoding.GetType()}");
            //AssertUtilities.WriteLine($"encoding: {encoding}");

            AssertUtilities.WriteLine();
            UANodeElementCollection descriptionNodeCollection = client.Browse(endpointDescriptor, dataTypeEncodingId,
                new UABrowseParameters(
                    UANodeClass.All, // may restrict more
                    EnumerableUtilities.One(UAReferenceTypeIds.HasDescription),
                    includeSubtypes: true));
            Assert.IsNotNull(descriptionNodeCollection);
            AssertUtilities.WriteLine($"descriptionNodeCollection.Count: {descriptionNodeCollection.Count}");
            Assert.AreEqual(1, descriptionNodeCollection.Count); // 0..1
            UANodeElement descriptionElement = descriptionNodeCollection[0];
            Assert.IsNotNull(descriptionElement);
            AssertUtilities.WriteLine($"descriptionElement: {descriptionElement}");
            UANodeId dataTypeDescriptionId = descriptionElement.NodeId;
            Assert.IsNotNull(dataTypeDescriptionId);
            AssertUtilities.WriteLine($"dataTypeDescriptionId: {dataTypeDescriptionId}");

            AssertUtilities.WriteLine();
            UANodeDescriptor dictionaryFragmentDescriptor = dataTypeDescriptionId.ToBrowsePath() 
                + UABrowsePathElement.HasProperty(UABrowseNames.DictionaryFragment);

            try
            {
                var dictionaryFragment = (byte[])client.ReadValue(endpointDescriptor, dictionaryFragmentDescriptor);
                Assert.IsNotNull(dictionaryFragment);
                AssertUtilities.WriteLine($"dictionaryFragment.GetType(): {dictionaryFragment.GetType()}");
                AssertUtilities.WriteLine($"dictionaryFragment: {dictionaryFragment}");
                // there seems to be no support for this in the UA SDK currently, and we do not use it here either
            }
            catch (UAException uaException) // DictionaryFragment is optional
            {
                // for performance, should not try/catch; use an overload that returns OperationResult derivative instead
                AssertUtilities.WriteLine(uaException.Message);
                // Got this - but not sure if it is always the case:
                // "An OPC-UA operation failure with error code -2140209152 (0x806F0000) occurred, originating from 'OpcLabs.EasyOpcUA'. The inner OPC-UA service exception with service result 'BadNoMatch' contains details about the problem."
            }

            // this value is a DataTypeSystem-specific "pointer" inside the DataTypeDictionary
            AssertUtilities.WriteLine();
            object dataTypeDescription = client.ReadValue(endpointDescriptor, dataTypeDescriptionId);
            Assert.IsNotNull(dataTypeDescription);
            AssertUtilities.WriteLine($"dataTypeDescription.GetType(): {dataTypeDescription.GetType()}");
            AssertUtilities.WriteLine($"dataTypeDescription: {dataTypeDescription}");

            // mimic Session.FindDataDictionary
            AssertUtilities.WriteLine();
            UANodeElementCollection dictionaryNodeCollection = client.Browse(endpointDescriptor, dataTypeDescriptionId,
                new UABrowseParameters(
                        UANodeClass.All,
                        EnumerableUtilities.One(UAReferenceTypeIds.HasComponent),
                        includeSubtypes: false) // really?
                    {BrowseDirections = UABrowseDirections.Inverse}
            );
            Assert.IsNotNull(dictionaryNodeCollection);
            AssertUtilities.WriteLine($"dictionaryNodeCollection.Count: {dictionaryNodeCollection.Count}");
            Assert.AreEqual(1, dictionaryNodeCollection.Count); // 0..1 (because we are following inverse HasComponent)
            UANodeElement dictionaryElement = dictionaryNodeCollection[0];
            Assert.IsNotNull(dictionaryElement);
            AssertUtilities.WriteLine($"dictionaryElement: {dictionaryElement}");
            UANodeId dictionaryId = dictionaryElement.NodeId;
            Assert.IsNotNull(dictionaryId);
            AssertUtilities.WriteLine($"dictionaryId: {dictionaryId}");

            AssertUtilities.WriteLine();
            object dictionary = client.ReadValue(endpointDescriptor, dictionaryId);
            Assert.IsNotNull(dictionary);
            AssertUtilities.WriteLine($"dictionary.GetType(): {dictionary.GetType()}");
            AssertUtilities.WriteLine($"dictionary: {dictionary}");
            var dictionaryArrayOfBytes = dictionary as byte[];
            Assert.IsNotNull(dictionaryArrayOfBytes);
            string dictionaryString = Encoding.UTF8.GetString(dictionaryArrayOfBytes);
            AssertUtilities.WriteLine($"dictionaryString: {dictionaryString}");

            //...
        }

        #endregion



        #region Private Properties

        [NotNull]
        private UALiveTestEnvironment TestEnvironment => UALiveTestEnvironment.Current;

        #endregion
    }
}
