﻿using System;
using System.Net.Sockets;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.UA
{
    public partial class EasyUAClientTest
    {
        [NotNull]
        static public UAEndpointDescriptor ClosedPortEndpointDescriptor => new UAEndpointDescriptor("opc.tcp://127.0.0.1:63999");

        static public void PrecheckClosedPort([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            try
            {
                using (new TcpClient(endpointDescriptor.Host ?? "", endpointDescriptor.Port))
#pragma warning disable 0642    // Possible mistaken empty statement
                    // ReSharper disable once EmptyEmbeddedStatement
                    /* do nothing*/
                    ;
#pragma warning restore 0642
            }
            catch (SocketException socketException)
            {
                if (socketException.SocketErrorCode != SocketError.ConnectionRefused)
                    Assert.Inconclusive("We need the connection be actively refused by the target machine");
            }
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void ClosedPortTest()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = ClosedPortEndpointDescriptor;
            PrecheckClosedPort(endpointDescriptor);

            var target = new EasyUAClient();
            try
            {
                try
                {
                    TestSection.Act();
                    target.Read(endpointDescriptor, "someNodeId");
                }
                finally
                {
                    TestSection.Assert();
                }
            }
            catch (UAException uaException)
            {
                Exception innerException = uaException.InnerException;
                Assert.IsNotNull(innerException);
                var serviceException = innerException as UAServiceException;
                Assert.IsNotNull(serviceException);
#if (NETFRAMEWORK)
                Assert.AreEqual(UACodeBits.BadTcpInternalError, serviceException.ServiceResult.StatusCode.CodeBits);
#else
                Assert.AreEqual(UACodeBits.BadNotConnected, serviceException.ServiceResult.StatusCode.CodeBits);
#endif
                return;
            }
            Assert.Fail("No exception thrown");
        }
    }
}
