﻿using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;

namespace OpcLabs.OpcInteroperabilityTest.UA.Servers
{
    [TestClass]
    // ReSharper disable once InconsistentNaming
    public class InductiveAutomation_IgnitionOpcUA_Test : UAInteroperabilityWorkshopTest
    {
        public override IEnumerable<UANodeDescriptor> ArrayReadableNodes => new UANodeDescriptor[]
        {
            "nsu=urn:inductiveautomation:ignition:opcua:tags ;ns=2;s=[default]/IntegerArray"
        };

        public override IEnumerable<UANodeDescriptor> DynamicNodes => new UANodeDescriptor[]
        {
            "nsu=urn:inductiveautomation:ignition:opcua:devices ;ns=1;s=[Generic Simulator]_Meta:Ramp/Ramp0",
            "nsu=urn:inductiveautomation:ignition:opcua:devices ;ns=1;s=[Generic Simulator]_Meta:Realistic/Realistic0",
            "nsu=urn:inductiveautomation:ignition:opcua:devices ;ns=1;s=[Generic Simulator]_Meta:Realistic/Realistic1"
        };

        public override IEnumerable<UAEndpointDescriptor> EndpointDescriptors => new UAEndpointDescriptor[] { "opc.tcp://192.168.1.24:62549/discovery" };

        public override int InterSessionDelay => 2000;

        public override string Password => "password";

        public override IEnumerable<string> ServerNotSupportedTests => base.ServerNotSupportedTests.Concat(new[]
        {
            nameof(Group001_Session_Test02_CreateSession_UaTcp_Basic128Rsa15),
            nameof(Group001_Session_Test09_CreateSession_Https_None),
            nameof(Group001_Session_Test10_ActivateSession_X509Certificate),
            nameof(Group003_AttributeServices_Test07_Write_Array_IndexRange),
            nameof(Group005_InformationModel_Test04_Browse_Read_SessionDiagnosticsArray),
            nameof(Group008_Eventing_Test01_Subscribe_Server),
            nameof(Group008_Eventing_Test04_EventNodes_Browse),
            nameof(Group011_AlarmCondition_Test01_Browse_Types),
            nameof(Group013_Discovery_Test03_FindServers_LDS_ServerPresence),
            //nameof(Group102_Session2_Test01_CreateSession_UaTcp_Basic256),
            //nameof(Group102_Session2_Test02_CreateSession_UaTcp_Basic256Sha256),
            //nameof(Group102_Session2_Test03_CreateSession_Https_Basic256Sha256)
        });

        public override IEnumerable<UANodeDescriptor> StaticNodes => new UANodeDescriptor[]
        {
            "nsu=urn:inductiveautomation:ignition:opcua:devices ;ns=1;s=[Generic Simulator]_Meta:ReadOnly/ReadOnlyInteger1",
            "nsu=urn:inductiveautomation:ignition:opcua:devices ;ns=1;s=[Generic Simulator]_Meta:ReadOnly/ReadOnlyFloat1"
        };

        public override IEnumerable<UANodeDescriptor> WriteableNodes => new UANodeDescriptor[]
        {
            "nsu=urn:inductiveautomation:ignition:opcua:devices ;ns=1;s=[Generic Simulator]_Meta:Writeable/WriteableInteger1",
            "nsu=urn:inductiveautomation:ignition:opcua:devices ;ns=1;s=[Generic Simulator]_Meta:Writeable/WriteableFloat1"
        };

        public override string UserName => "opcuauser";
    }
}
