﻿
using System.Diagnostics;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Discovery;
using OpcLabs.UnitTestLib;

namespace OpcLabs.OpcInteroperabilityTest.UA
{
    partial class UAInteroperabilityWorkshopTest
    {
        #region Test Methods

        [TestMethod]
        public void Group013_Discovery_Header____________________()
        {
        }

        /// <summary>
        /// Call FindServers() on a remote LDS and specify a ServerURI to get a specific Server application.
        /// </summary>
        [TestMethod]
        public void Group013_Discovery_Test01_FindServers_LDS_ServerUri()
        {
            ClientNotSupported();   // IMPROVE: have this
        }

        /// <summary>
        /// Call GetEndpoints() on the Server application that is the testee. Specify the desired transport profile in the request. Verify that only endpoints are returned that include this profile.
        /// </summary>
        [TestMethod]
        public void Group013_Discovery_Test02_GetEndpoints_Server()
        {
            ClientNotSupported();   // IMPROVE: have this
        }

        /// <summary>
        /// Call FindServers() on a remote LDS to get the list of available Server applications. Verify that the Server application that is the testee is in this list.
        /// </summary>
        [TestMethod]
        public void Group013_Discovery_Test03_FindServers_LDS_ServerPresence()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;
            AssertUtilities.WriteLine($"Endpoint descriptor: {endpointDescriptor}");
            string discoveryHost = endpointDescriptor.Host;
            Debug.Assert(discoveryHost != null);

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
            {
                UADiscoveryElementCollection discoveryElementCollection = client.DiscoverLocalServers(discoveryHost);

                AssertUtilities.WriteLine();
                AssertUtilities.WriteLine("Application elements:");
                foreach (UADiscoveryElement discoveryElement in discoveryElementCollection)
                    // ReSharper disable once PossibleNullReferenceException
                    AssertUtilities.WriteLine($"{discoveryElement.ApplicationName}: {discoveryElement.DiscoveryUriString}");
                AssertUtilities.WriteLine($"Element count: {discoveryElementCollection.Count}");

                TestSection.Assert();

                string expectedEndpointDescriptorUrlString = endpointDescriptor.UrlString;
                AssertUtilities.WriteLine($"Expected endpoint URL string: {expectedEndpointDescriptorUrlString}");

                // ReSharper disable once PossibleNullReferenceException
                Assert.IsTrue(discoveryElementCollection.Any(element => 
                        (element.DiscoveryUriString.ToLowerInvariant() == expectedEndpointDescriptorUrlString.ToLowerInvariant())),
                    "URI not found in the returned list");
            }
        }

        /// <summary>
        /// Call GetEndpoints() on the Server application that is the testee.
        /// </summary>
        [TestMethod]
        public void Group013_Discovery_Test04_GetEndpoints_Server()
        {
            TestSection.Arrange();
            PrecheckTestMethod();

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
                CreateSession(client, EndpointDescriptorCommon);
            // The ability to create a session in this way is a proof of test success here
        }

        // IMPROVE: have more

        #endregion
    }
}
