﻿
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.CompilerServices;
using JetBrains.Annotations;
using OpcLabs.BaseLib.Utilities;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Discovery;
using OpcLabs.EasyOpc.UA.Extensions;
using OpcLabs.EasyOpc.UA.Implementation;
using OpcLabs.EasyOpc.UA.Internal;
using OpcLabs.UnitTestLib;

namespace OpcLabs.OpcInteroperabilityTest.UA
{
    public abstract partial class UAInteroperabilityWorkshopTest
    {
        #region Protected Constructors

        protected  UAInteroperabilityWorkshopTest()
        {
            BaseLib.Diagnostics.Switches.CallDiagnostics.Display.EnabledSwitch.Enabled = false;
            Instrumentation.Toolkit.TraceSdkSwitch.Enabled = false;
            EasyUAClient.SharedParameters.EngineParameters.CertificateAcceptancePolicy.AcceptAnyCertificate = true;

            _clientNotSupportedTestSet = new Lazy<HashSet<string>>(() => new HashSet<string>(ClientNotSupportedTests));
            _serverNotSupportedTestSet = new Lazy<HashSet<string>>(() => new HashSet<string>(ServerNotSupportedTests));
        }

        #endregion



        #region Public Properties

        [CanBeNull]
        public UANodeDescriptor ArrayReadableNode1 => ArrayReadableNodes.FirstOrDefault();

        [NotNull]
        public virtual IEnumerable<UANodeDescriptor> ArrayReadableNodes => Enumerable.Empty<UANodeDescriptor>();

        [CanBeNull]
        public UANodeDescriptor ArrayWriteableNode1 => ArrayWriteableNodes.FirstOrDefault();

        [NotNull]
        public virtual IEnumerable<UANodeDescriptor> ArrayWriteableNodes => Enumerable.Empty<UANodeDescriptor>();

        [NotNull]
        public virtual IEnumerable<string> ClientNotSupportedTests => Enumerable.Empty<string>();

        [CanBeNull]
        public UAEndpointDescriptor ConfiguredEndpointDescriptorUATcp
        {
            get
            {
                UAEndpointDescriptor endpointDescriptorUATcp = EndpointDescriptorUATcp;
                if (endpointDescriptorUATcp == null)
                    return null;
                // ReSharper disable once PossibleNullReferenceException
                return _serverNotSupportedTestSet.Value.Contains(nameof(Group001_Session_Test04_ActivateSession_UserIdentityAnonymous)) 
                    // ReSharper disable once AssignNullToNotNullAttribute
                    ? new UAEndpointDescriptor(endpointDescriptorUATcp).WithUserNameIdentity(UserName, Password)
                    : endpointDescriptorUATcp;
            }
        }

        [CanBeNull]
        public UANodeDescriptor DeadbandNode1 => DeadbandNodes.FirstOrDefault();

        [CanBeNull]
        public UANodeDescriptor DeadbandNode2 => DeadbandNodes.ElementAtOrDefault(1);

        [NotNull]
        public virtual IEnumerable<UANodeDescriptor> DeadbandNodes => DynamicNodes;

        public int DeadbandSamplingInterval1 => DeadbandSamplingIntervals.ElementAt(0);

        public int DeadbandSamplingInterval2 => DeadbandSamplingIntervals.ElementAt(1);

        [NotNull]
        public virtual IEnumerable<int> DeadbandSamplingIntervals => new[] { 1000 };

        [CanBeNull]
        public UANodeDescriptor DynamicNode1 => DynamicNodes.FirstOrDefault();

        [CanBeNull]
        public UANodeDescriptor DynamicNode2 => DynamicNodes.ElementAtOrDefault(1);

        [NotNull]
        public virtual IEnumerable<UANodeDescriptor> DynamicNodes => Enumerable.Empty<UANodeDescriptor>();

        public int DynamicSamplingInterval1 => DynamicSamplingIntervals.ElementAt(0);

        public int DynamicSamplingInterval2 => DynamicSamplingIntervals.ElementAt(1);

        [NotNull]
        public virtual IEnumerable<int> DynamicSamplingIntervals => new[] { 2000, 5000 };

        [NotNull]
        // ReSharper disable once AssignNullToNotNullAttribute
        public virtual UAEndpointDescriptor EndpointDescriptorCommon
        {
            get
            {
                UAEndpointDescriptor endpointDescriptorUATcp = ConfiguredEndpointDescriptorUATcp;
                if (endpointDescriptorUATcp != null)
                    return endpointDescriptorUATcp;
                UAEndpointDescriptor endpointDescriptorHttps = EndpointDescriptorHttps;
                // ReSharper disable once ConvertIfStatementToReturnStatement
                if (endpointDescriptorHttps != null)
                    return endpointDescriptorHttps;
                
                // ReSharper disable once AssignNullToNotNullAttribute
                return EndpointDescriptors.First();
            }
        }

        [CanBeNull]
        public virtual UAEndpointDescriptor EndpointDescriptorHttps => EndpointDescriptors.FirstOrDefault(descriptor =>
            // ReSharper disable once PossibleNullReferenceException
            UriUtilities.HasScheme(descriptor.UrlString, UASchemes.Https));

        [NotNull]
        public abstract IEnumerable<UAEndpointDescriptor> EndpointDescriptors { get; }

        [CanBeNull]
        public virtual UAEndpointDescriptor EndpointDescriptorUATcp => EndpointDescriptors.FirstOrDefault(descriptor =>
            // ReSharper disable once PossibleNullReferenceException
            UriUtilities.HasScheme(descriptor.UrlString, UASchemes.OpcTcp));

        [NotNull]
        // ReSharper disable once AssignNullToNotNullAttribute
        public virtual UAEndpointDescriptor EndpointDescriptorUserNameAndPassword => EndpointDescriptorCommon;

        public virtual int InterSessionDelay => 0;

        public virtual int MaximumSearchNodes => 100;

        [NotNull]
        public virtual string Password => string.Empty;

        [NotNull]
        public virtual IEnumerable<UANodeDescriptor> ReadableNodes => Enumerable.Empty<UANodeDescriptor>();

        [NotNull]
        public virtual IEnumerable<string> ServerNotSupportedTests => Enumerable.Empty<string>();

        [CanBeNull]
        public UANodeDescriptor StaticNode1 => StaticNodes.FirstOrDefault();

        [CanBeNull]
        public UANodeDescriptor StaticNode2 => StaticNodes.ElementAtOrDefault(1);

        [NotNull]
        public virtual IEnumerable<UANodeDescriptor> StaticNodes => Enumerable.Empty<UANodeDescriptor>();

        public virtual bool SupportsJustOneSubscription => false;

        [CanBeNull]
        public virtual string UserName => null;

        [CanBeNull]
        public virtual string X509CertificateFileName => null;

        [NotNull]
        public virtual string X509CertificatePassword => string.Empty;

        [CanBeNull]
        public UANodeDescriptor WriteableNode1 => WriteableNodes.FirstOrDefault();

        [NotNull]
        public virtual IEnumerable<UANodeDescriptor> WriteableNodes => Enumerable.Empty<UANodeDescriptor>();

        public virtual bool WriteWithUserNameIdentity => false;

        #endregion



        #region Protected Methods

        [ContractAnnotation("=> halt")]
        static protected void ClientNotSupported()
        {
            AssertUtilities.Inconclusive("Client not supported");
        }

        protected void PrecheckTestMethod([CallerMemberName] [NotNull] string methodName = "")
        {
            // ReSharper disable once PossibleNullReferenceException
            if (_clientNotSupportedTestSet.Value.Contains(methodName))
                ClientNotSupported();
            // ReSharper disable once PossibleNullReferenceException
            if (_serverNotSupportedTestSet.Value.Contains(methodName))
                ServerNotSupported();
        }

        [ContractAnnotation("=> halt")]
        static protected void ServerNotSupported()
        {
            AssertUtilities.Inconclusive("Server not supported");
        }

        #endregion



        #region Private Fields

        [NotNull]
        private readonly Lazy<HashSet<string>> _clientNotSupportedTestSet;

        [NotNull]
        private readonly Lazy<HashSet<string>> _serverNotSupportedTestSet;

        #endregion
    }
}
